

res = {}
with open('programs.md', 'r') as f:
    for line in f:
        if line.startswith('#'):
            continue
        try:
            pid, name = line[38:].split('    ')
            pid = int(pid)
            name = name.strip()
            if name.find(' ') != -1:
                name = name.replace(' ', '/')
            res[pid] = name
        except:
            print(line[34:])
            input(line)
            continue

print(repr(res))



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import enum
import copy

from anfs.protocol.rpc import RPC
from anfs.protocol.portmap import Portmap
from anfs.protocol.mount import Mount
from anfs.protocol.rpc.common.target import RPCTarget
from anfs.protocol.rpc.messages import AUTH_SYS
from anfs.protocol.nfs3.client import NFSv3Client
from asyauth.common.credentials import UniCredential

class NFS3ConnectionFactory:	
	def __init__(self, credential:UniCredential = None, target:RPCTarget = None ):
		self.credential = credential
		self.target = target

	@staticmethod
	def from_url(connection_url):
		target = RPCTarget.from_url(connection_url)
		credential = AUTH_SYS()
		#credential = UniCredential.from_url(connection_url)
		return NFS3ConnectionFactory(credential, target)

	def get_credential(self) -> UniCredential:
		"""
		Creates a credential object
		
		:return: Credential object
		:rtype: :class:`UniCredential`
		"""
		return copy.deepcopy(self.credential)

	def get_target(self) -> RPCTarget:
		"""
		Creates a target object
		
		:return: Target object
		:rtype: :class:`RPCTarget`
		"""
		return copy.deepcopy(self.target)
	
	def get_portmap(self) -> Portmap:
		"""
		Creates a portmap object
		"""

		target = self.get_target()
		target.port = 111
		cred = self.get_credential()
		return Portmap(target, cred)
	
	def get_mount(self):
		"""
		Creates a mount object
		
		:return: Mount object
		:rtype: :class:`Mount`
		"""
		
		target = self.get_target()
		cred = self.get_credential()
		return Mount(target, cred)

	def get_client(self, root_handle) -> NFSv3Client:
		"""
		Creates a client that can be used to interface with the server
		
		:return: NFSv3 client
		:rtype: :class:`NFSv3Client`
		"""
		cred = self.get_credential()
		target = self.get_target()
		return NFSv3Client(root_handle, target, cred)


	def get_connection(self) -> RPC:
		"""
		Creates a connection that can be used to interface with the server
		
		:return: RPC connection
		:rtype: :class:`RPC`
		"""
		cred = self.get_credential()
		target = self.get_target()
		return RPC(target, cred)
		
	def __str__(self):
		t = '==== NFSConnectionFactory ====\r\n'
		for k in self.__dict__:
			val = self.__dict__[k]
			if isinstance(val, enum.IntFlag):
				val = val
			elif isinstance(val, enum.Enum):
				val = val.name
			
			t += '%s: %s\r\n' % (k, str(val))
			
		return t





import io
import os

MNTPATHLEN = 1024
MNTNAMLEN = 255
FHSIZE = 32

def read_str(buf:io.BytesIO, encoding:str='utf-8'):
    sl = int.from_bytes(buf.read(4), byteorder='big', signed=False)
    data = buf.read(sl).decode(encoding)
    if sl%4 != 0:
        buf.read(4 - sl%4)
    return data

def pack_str(data:str, encoding:str='utf-8'):
    data = data.encode(encoding)
    padding =  b'' if len(data) % 4 == 0 else b'\x00' * (4 - len(data)%4)
    return len(data).to_bytes(4, byteorder='big', signed=False) + data + padding

class export:
    def __init__(self):
        self.filesys = None
        self.groups = []
    
    @staticmethod
    def from_bytes(data: bytes):
        return export.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = export()
        result.filesys = read_str(buff)
        while buff.read(4) != b'\x00\x00\x00\x00':
            groupname = read_str(buff)
            result.groups.append(groupname)
            
        return result
    
    def __str__(self):
        return 'Filesystem: %s, Groups: %s' % (self.filesys, self.groups)
    
class fhstatus:
    def __init__(self):
        self.status = None
        self.fhandle = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return fhstatus.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = fhstatus()
        result.status = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        if result.status == 0:
            hszie = int.from_bytes(buff.read(4), byteorder='big', signed=False)
            result.fhandle = buff.read(hszie)
        else:
            raise Exception('fhstatus error: %s' % os.strerror(result.status))
        return result
    
    def __str__(self):
        return 'Status: %s' % self.status

class mountpoint:
    def __init__(self):
        self.hostname = None
        self.directory = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return mountpoint.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = mountpoint()
        result.hostname = read_str(buff)
        result.directory = read_str(buff)
        return result
    
    def __str__(self):
        return 'Hostname: %s, Directory: %s' % (self.hostname, self.directory)


import copy
from asysocks.unicomm.common.target import UniTarget, UniProto
from anfs.protocol.portmap.messages import ipproto
from urllib.parse import urlparse, parse_qs
from asysocks.unicomm.utils.paramprocessor import str_one, int_one, bool_one

rpctarget_url_params = {
	'fragsize': int_one,
}

class RPCTarget(UniTarget):
	def __init__(self, ip, port = 0, protocol = UniProto.CLIENT_TCP, proxies = None, timeout = 10, dns:str=None, dc_ip:str = None, domain:str = None, hostname:str = None, ssl_ctx = None, use_privileged_source_port:bool=False, fragsize = 10*1024):
		UniTarget.__init__(self, ip, port, protocol, timeout, hostname = hostname, ssl_ctx= ssl_ctx, proxies = proxies, domain = domain, dc_ip = dc_ip, dns=dns, use_privileged_source_port=use_privileged_source_port)
		self.fragsize = fragsize
	
	def get_rpcprotocol(self):
		if self.protocol in [UniProto.CLIENT_TCP, UniProto.SERVER_TCP, UniProto.CLIENT_SSL_TCP, UniProto.SERVER_SSL_TCP]:
			return ipproto.IPPROTO_TCP.value
		if self.protocol in [UniProto.CLIENT_UDP, UniProto.SERVER_UDP]:
			return ipproto.IPPROTO_UDP.value
		raise Exception('Unknown protocol')

	def get_newtarget(self, ip, port, hostname = None):
		t = copy.deepcopy(self)
		t.ip = ip
		t.port = port
		t.hostname = hostname
		return t

	@staticmethod
	def from_url(connection_url:str):
		url_e = urlparse(connection_url)
		schemes = []
		for item in url_e.scheme.upper().split('+'):
			schemes.append(item.replace('-','_'))

		port = 0
		if url_e.port:
			port = url_e.port
		if port is None:
			raise Exception('Port must be provided!')
		
		path = None
		if url_e.path not in ['/', '', None]:
			path = url_e.path
		
		protocol = UniProto.CLIENT_TCP
		unitarget, extraparams = UniTarget.from_url(connection_url, protocol, port, rpctarget_url_params)
		fragsize = extraparams['fragsize'] if extraparams['fragsize'] is not None else 10*1024

		target = RPCTarget(
			unitarget.ip, 
			port = unitarget.port, 
			protocol = unitarget.protocol, 
			proxies = unitarget.proxies, 
			timeout = unitarget.timeout, 
			dns = unitarget.dns, 
			dc_ip = unitarget.dc_ip, 
			domain = unitarget.domain, 
			hostname = unitarget.hostname,
			ssl_ctx = unitarget.ssl_ctx,
			fragsize = fragsize,
			use_privileged_source_port = unitarget.use_privileged_source_port
		)
		return target


from anfs.protocol.rpc import RPC
from anfs.protocol.portmap.messages import mapping, call_args, call_result
import traceback
import struct
import io
import datetime

# https://www.rfc-editor.org/rfc/rfc1833.html
# TODO: Implement the rest of the portmap protocol, as many functions missing

class Portmap:
    def __init__(self, target, credential = None):
        self.target = target
        self.credential = credential
        self.rpc = None
    
    async def __aenter__(self):
        _, err = await self.connect()
        if err is not None:
            raise err
        return self
    
    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()
    
    async def connect(self):
        self.rpc = RPC(self.target, self.credential)
        return await self.rpc.connect(100000, 4)
    
    async def disconnect(self):
        if self.rpc is not None:
            return await self.rpc.disconnect()
    
    async def dump(self):
        """Fetches all the portmap entries from the server"""
        try:
            data = struct.pack('!LL', 2, 4)
            pm_entries = []
            pm_entries_buff, err = await self.rpc.call(100000, 2, 4, data)
            if err is not None:
                raise err
            if len(pm_entries_buff) < 4:
                return pm_entries
            
            pm_entries_buff = io.BytesIO(pm_entries_buff)
            while pm_entries_buff.read(4) == b'\x00\x00\x00\x01':
                pm_entries.append(mapping.from_buffer(pm_entries_buff))
            return pm_entries, None
        except Exception as e:
            return None, e

    async def getport(self, program, version, protocol):
        """Fetches the port for a specific program and version"""
        try:
            data = mapping()
            data.program = program
            data.version = version
            data.protocol = protocol
            data.port = 0
            data = data.to_bytes()
            
            port_buff, err = await self.rpc.call(100000, 2, 3, data)
            if err is not None:
                raise err
            if len(port_buff) < 4:
                return None
            return int.from_bytes(port_buff, byteorder='big', signed=False), None
        except Exception as e:
            return None, e
    
    async def null(self):
        """Null call"""
        return await self.rpc.call(100000, 2, 0, b'')
    
    async def callit(self, program, version, procedure, data):
        """Call a specific procedure on a program"""
        try:
            req = call_args()
            req.program = program
            req.version = version
            req.procedure = procedure
            req.data = data
            req = req.to_bytes()
            
            rep, err =  await self.rpc.call(100000, 2, 5, req)
            if err is not None:
                raise err
            return call_result.from_bytes(rep), None
        except Exception as e:
            return None, e
    
    async def gettime(self):
        """Get the server time"""
        try:
            tdata, err = await self.rpc.call(100000, 4, 6, b'')
            if err is not None:
                raise err
            curtime = datetime.datetime.fromtimestamp(int.from_bytes(tdata, byteorder='big', signed=False))
            return curtime, None
        except Exception as e:
            return None, e
        

async def amain():
    from anfs.protocol.rpc.common.target import RPCTarget

    ip = '192.168.30.10'
    target = RPCTarget(ip, port = 111)
    portmap = Portmap(target)
    try:
        _, err = await portmap.connect()
        if err is not None:
            raise err
        pm, err = await portmap.dump()
        for p in pm:
            print(str(p))
        #port = await portmap.getport(100000, 2, 6)
        #print(port)
        #curtime = await portmap.gettime()
        #print(curtime)
    except Exception as e:
        traceback.print_exc()
def main():
    import asyncio
    asyncio.run(amain())

if __name__ == '__main__':
    main()


from anfs.protocol.nfs3 import messages as const


class specdata3:
    # XDR definition:
    # struct specdata3 {
    #     uint32 major;
    #     uint32 minor;
    # };
    def __init__(self, major=None, minor=None):
        self.major = major
        self.minor = minor

    def __repr__(self):
        out = []
        if self.major is not None:
            out += ['major=%s' % repr(self.major)]
        if self.minor is not None:
            out += ['minor=%s' % repr(self.minor)]
        return 'specdata3(%s)' % ', '.join(out)


class nfs_fh3:
    # XDR definition:
    # struct nfs_fh3 {
    #     opaque data<NFS3_FHSIZE>;
    # };
    def __init__(self, data=None):
        self.data = data

    def __repr__(self):
        out = []
        if self.data is not None:
            out += ['data=%s' % repr(self.data)]
        return 'nfs_fh3(%s)' % ', '.join(out)


class nfstime3:
    # XDR definition:
    # struct nfstime3 {
    #     uint32 seconds;
    #     uint32 nseconds;
    # };
    def __init__(self, seconds=None, nseconds=None):
        self.seconds = seconds
        self.nseconds = nseconds

    def __repr__(self):
        out = []
        if self.seconds is not None:
            out += ['seconds=%s' % repr(self.seconds)]
        if self.nseconds is not None:
            out += ['nseconds=%s' % repr(self.nseconds)]
        return 'nfstime3(%s)' % ', '.join(out)


class fattr3:
    # XDR definition:
    # struct fattr3 {
    #     ftype3 type;
    #     uint32 mode;
    #     uint32 nlink;
    #     uint32 uid;
    #     uint32 gid;
    #     uint64 size;
    #     uint64 used;
    #     specdata3 rdev;
    #     uint64 fsid;
    #     uint64 fileid;
    #     nfstime3 atime;
    #     nfstime3 mtime;
    #     nfstime3 ctime;
    # };
    def __init__(self, type=None, mode=None, nlink=None, uid=None, gid=None, size=None, used=None, rdev=None,
                 fsid=None, fileid=None, atime=None, mtime=None, ctime=None):
        self.type = type
        self.mode = mode
        self.nlink = nlink
        self.uid = uid
        self.gid = gid
        self.size = size
        self.used = used
        self.rdev = rdev
        self.fsid = fsid
        self.fileid = fileid
        self.atime = atime
        self.mtime = mtime
        self.ctime = ctime

    def __repr__(self):
        out = []
        if self.type is not None:
            out += ['type=%s' % const.FTYPE3.get(self.type, self.type)]
        if self.mode is not None:
            out += ['mode=%s' % repr(self.mode)]
        if self.nlink is not None:
            out += ['nlink=%s' % repr(self.nlink)]
        if self.uid is not None:
            out += ['uid=%s' % repr(self.uid)]
        if self.gid is not None:
            out += ['gid=%s' % repr(self.gid)]
        if self.size is not None:
            out += ['size=%s' % repr(self.size)]
        if self.used is not None:
            out += ['used=%s' % repr(self.used)]
        if self.rdev is not None:
            out += ['rdev=%s' % repr(self.rdev)]
        if self.fsid is not None:
            out += ['fsid=%s' % repr(self.fsid)]
        if self.fileid is not None:
            out += ['fileid=%s' % repr(self.fileid)]
        if self.atime is not None:
            out += ['atime=%s' % repr(self.atime)]
        if self.mtime is not None:
            out += ['mtime=%s' % repr(self.mtime)]
        if self.ctime is not None:
            out += ['ctime=%s' % repr(self.ctime)]
        return 'fattr3(%s)' % ', '.join(out)


class post_op_attr:
    # XDR definition:
    # union post_op_attr switch(bool present) {
    #     case TRUE:
    #         fattr3 attributes;
    #     case FALSE:
    #         void;
    # };
    def __init__(self, present=None, attributes=None):
        self.present = present
        self.attributes = attributes

    def __repr__(self):
        out = []
        if self.present is not None:
            out += ['present=%s' % repr(self.present)]
        if self.attributes is not None:
            out += ['attributes=%s' % repr(self.attributes)]
        return 'post_op_attr(%s)' % ', '.join(out)


class wcc_attr:
    # XDR definition:
    # struct wcc_attr {
    #     uint64 size;
    #     nfstime3 mtime;
    #     nfstime3 ctime;
    # };
    def __init__(self, size=None, mtime=None, ctime=None):
        self.size = size
        self.mtime = mtime
        self.ctime = ctime

    def __repr__(self):
        out = []
        if self.size is not None:
            out += ['size=%s' % repr(self.size)]
        if self.mtime is not None:
            out += ['mtime=%s' % repr(self.mtime)]
        if self.ctime is not None:
            out += ['ctime=%s' % repr(self.ctime)]
        return 'wcc_attr(%s)' % ', '.join(out)


class pre_op_attr:
    # XDR definition:
    # union pre_op_attr switch(bool present) {
    #     case TRUE:
    #         wcc_attr attributes;
    #     case FALSE:
    #         void;
    # };
    def __init__(self, present=None, attributes=None):
        self.present = present
        self.attributes = attributes

    def __repr__(self):
        out = []
        if self.present is not None:
            out += ['present=%s' % repr(self.present)]
        if self.attributes is not None:
            out += ['attributes=%s' % repr(self.attributes)]
        return 'pre_op_attr(%s)' % ', '.join(out)


class wcc_data:
    # XDR definition:
    # struct wcc_data {
    #     pre_op_attr before;
    #     post_op_attr after;
    # };
    def __init__(self, before=None, after=None):
        self.before = before
        self.after = after

    def __repr__(self):
        out = []
        if self.before is not None:
            out += ['before=%s' % repr(self.before)]
        if self.after is not None:
            out += ['after=%s' % repr(self.after)]
        return 'wcc_data(%s)' % ', '.join(out)


class post_op_fh3:
    # XDR definition:
    # union post_op_fh3 switch(bool present) {
    #     case TRUE:
    #         nfs_fh3 handle;
    #     case FALSE:
    #         void;
    # };
    def __init__(self, present=None, handle=None):
        self.present = present
        self.handle = handle

    def __repr__(self):
        out = []
        if self.present is not None:
            out += ['present=%s' % repr(self.present)]
        if self.handle is not None:
            out += ['handle=%s' % repr(self.handle)]
        return 'post_op_fh3(%s)' % ', '.join(out)


class set_uint32:
    # XDR definition:
    # union set_uint32 switch(bool set) {
    #     case TRUE:
    #         uint32 val;
    #     default:
    #         void;
    # };
    def __init__(self, set=None, val=None):
        self.set = set
        self.val = val

    def __repr__(self):
        out = []
        if self.set is not None:
            out += ['set=%s' % repr(self.set)]
        if self.val is not None:
            out += ['val=%s' % repr(self.val)]
        return 'set_uint32(%s)' % ', '.join(out)


class set_uint64:
    # XDR definition:
    # union set_uint64 switch(bool set) {
    #     case TRUE:
    #         uint64 val;
    #     default:
    #         void;
    # };
    def __init__(self, set=None, val=None):
        self.set = set
        self.val = val

    def __repr__(self):
        out = []
        if self.set is not None:
            out += ['set=%s' % repr(self.set)]
        if self.val is not None:
            out += ['val=%s' % repr(self.val)]
        return 'set_uint64(%s)' % ', '.join(out)


class set_time:
    # XDR definition:
    # union set_time switch(time_how set) {
    #     case SET_TO_CLIENT_TIME:
    #         nfstime3 time;
    #     default:
    #         void;
    # };
    def __init__(self, set=None, time=None):
        self.set = set
        self.time = time

    def __repr__(self):
        out = []
        if self.set is not None:
            out += ['set=%s' % const.time_how.get(self.set, self.set)]
        if self.time is not None:
            out += ['time=%s' % repr(self.time)]
        return 'set_time(%s)' % ', '.join(out)


class sattr3:
    # XDR definition:
    # struct sattr3 {
    #     set_uint32 mode;
    #     set_uint32 uid;
    #     set_uint32 gid;
    #     set_uint64 size;
    #     set_time atime;
    #     set_time mtime;
    # };
    def __init__(self, mode=None, uid=None, gid=None, size=None, atime=None, mtime=None):
        self.mode = mode
        self.uid = uid
        self.gid = gid
        self.size = size
        self.atime = atime
        self.mtime = mtime

    def __repr__(self):
        out = []
        if self.mode is not None:
            out += ['mode=%s' % repr(self.mode)]
        if self.uid is not None:
            out += ['uid=%s' % repr(self.uid)]
        if self.gid is not None:
            out += ['gid=%s' % repr(self.gid)]
        if self.size is not None:
            out += ['size=%s' % repr(self.size)]
        if self.atime is not None:
            out += ['atime=%s' % repr(self.atime)]
        if self.mtime is not None:
            out += ['mtime=%s' % repr(self.mtime)]
        return 'sattr3(%s)' % ', '.join(out)


class diropargs3:
    # XDR definition:
    # struct diropargs3 {
    #     nfs_fh3 dir;
    #     filename3 name;
    # };
    def __init__(self, dir=None, name=None):
        self.dir = dir
        self.name = name

    def __repr__(self):
        out = []
        if self.dir is not None:
            out += ['dir=%s' % repr(self.dir)]
        if self.name is not None:
            out += ['name=%s' % repr(self.name)]
        return 'diropargs3(%s)' % ', '.join(out)


class diropres3ok:
    # XDR definition:
    # struct diropres3ok {
    #     post_op_fh3 obj;
    #     post_op_attr obj_attributes;
    #     wcc_data dir_wcc;
    # };
    def __init__(self, obj=None, obj_attributes=None, dir_wcc=None):
        self.obj = obj
        self.obj_attributes = obj_attributes
        self.dir_wcc = dir_wcc

    def __repr__(self):
        out = []
        if self.obj is not None:
            out += ['obj=%s' % repr(self.obj)]
        if self.obj_attributes is not None:
            out += ['obj_attributes=%s' % repr(self.obj_attributes)]
        if self.dir_wcc is not None:
            out += ['dir_wcc=%s' % repr(self.dir_wcc)]
        return 'diropres3ok(%s)' % ', '.join(out)


class diropres3:
    # XDR definition:
    # union diropres3 switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         diropres3ok resok;
    #     default:
    #         wcc_data resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'diropres3(%s)' % ', '.join(out)


class wccstat3:
    # XDR definition:
    # union wccstat3 switch(nfsstat3 status) {
    #     case -1:
    #         void;
    #     default:
    #         wcc_data wcc;
    # };
    def __init__(self, status=None):
        self.status = status

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        return 'wccstat3(%s)' % ', '.join(out)


class getattr3res:
    # XDR definition:
    # union getattr3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         fattr3 attributes;
    #     default:
    #         void;
    # };
    def __init__(self, status=None, attributes=None):
        self.status = status
        self.attributes = attributes

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.attributes is not None:
            out += ['attributes=%s' % repr(self.attributes)]
        return 'getattr3res(%s)' % ', '.join(out)


class sattrguard3:
    # XDR definition:
    # union sattrguard3 switch(bool check) {
    #     case TRUE:
    #         nfstime3 ctime;
    #     case FALSE:
    #         void;
    # };
    def __init__(self, check=None, ctime=None):
        self.check = check
        self.ctime = ctime

    def __repr__(self):
        out = []
        if self.check is not None:
            out += ['check=%s' % repr(self.check)]
        if self.ctime is not None:
            out += ['ctime=%s' % repr(self.ctime)]
        return 'sattrguard3(%s)' % ', '.join(out)


class setattr3args:
    # XDR definition:
    # struct setattr3args {
    #     nfs_fh3 object;
    #     sattr3 new_attributes;
    #     sattrguard3 guard;
    # };
    def __init__(self, object=None, new_attributes=None, guard=None):
        self.object = object
        self.new_attributes = new_attributes
        self.guard = guard

    def __repr__(self):
        out = []
        if self.object is not None:
            out += ['object=%s' % repr(self.object)]
        if self.new_attributes is not None:
            out += ['new_attributes=%s' % repr(self.new_attributes)]
        if self.guard is not None:
            out += ['guard=%s' % repr(self.guard)]
        return 'setattr3args(%s)' % ', '.join(out)


class wcc_data3res:
    def __init__(self, category, status=None, wcc_data=None):
        self.category = str(category) or ''
        self.status = status
        self.wcc_data = wcc_data

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.wcc_data is not None:
            out += ['wcc_data=%s' % repr(self.wcc_data)]
        return self.category + '(%s)' % ', '.join(out)


class lookup3resok:
    # XDR definition:
    # struct lookup3resok {
    #     nfs_fh3 object;
    #     post_op_attr obj_attributes;
    #     post_op_attr dir_attributes;
    # };
    def __init__(self, object=None, obj_attributes=None, dir_attributes=None):
        self.object = object
        self.obj_attributes = obj_attributes
        self.dir_attributes = dir_attributes

    def __repr__(self):
        out = []
        if self.object is not None:
            out += ['object=%s' % repr(self.object)]
        if self.obj_attributes is not None:
            out += ['obj_attributes=%s' % repr(self.obj_attributes)]
        if self.dir_attributes is not None:
            out += ['dir_attributes=%s' % repr(self.dir_attributes)]
        return 'lookup3resok(%s)' % ', '.join(out)


class lookup3res:
    # XDR definition:
    # union lookup3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         lookup3resok resok;
    #     default:
    #         post_op_attr resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'lookup3res(%s)' % ', '.join(out)


class access3args:
    # XDR definition:
    # struct access3args {
    #     nfs_fh3 object;
    #     uint32 access;
    # };
    def __init__(self, object=None, access=None):
        self.object = object
        self.access = access

    def __repr__(self):
        out = []
        if self.object is not None:
            out += ['object=%s' % repr(self.object)]
        if self.access is not None:
            out += ['access=%s' % repr(self.access)]
        return 'access3args(%s)' % ', '.join(out)


class access3resok:
    # XDR definition:
    # struct access3resok {
    #     post_op_attr obj_attributes;
    #     uint32 access;
    # };
    def __init__(self, obj_attributes=None, access=None):
        self.obj_attributes = obj_attributes
        self.access = access

    def __repr__(self):
        out = []
        if self.obj_attributes is not None:
            out += ['obj_attributes=%s' % repr(self.obj_attributes)]
        if self.access is not None:
            out += ['access=%s' % repr(self.access)]
        return 'access3resok(%s)' % ', '.join(out)


class access3res:
    # XDR definition:
    # union access3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         access3resok resok;
    #     default:
    #         post_op_attr resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'access3res(%s)' % ', '.join(out)


class readlink3resok:
    # XDR definition:
    # struct readlink3resok {
    #     post_op_attr symlink_attributes;
    #     nfspath3 data;
    # };
    def __init__(self, symlink_attributes=None, data=None):
        self.symlink_attributes = symlink_attributes
        self.data = data

    def __repr__(self):
        out = []
        if self.symlink_attributes is not None:
            out += ['symlink_attributes=%s' % repr(self.symlink_attributes)]
        if self.data is not None:
            out += ['data=%s' % repr(self.data)]
        return 'readlink3resok(%s)' % ', '.join(out)


class readlink3res:
    # XDR definition:
    # union readlink3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         readlink3resok resok;
    #     default:
    #         post_op_attr resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'readlink3res(%s)' % ', '.join(out)


class read3args:
    # XDR definition:
    # struct read3args {
    #     nfs_fh3 file;
    #     uint64 offset;
    #     uint32 count;
    # };
    def __init__(self, file=None, offset=None, count=None):
        self.file = file
        self.offset = offset
        self.count = count

    def __repr__(self):
        out = []
        if self.file is not None:
            out += ['file=%s' % repr(self.file)]
        if self.offset is not None:
            out += ['offset=%s' % repr(self.offset)]
        if self.count is not None:
            out += ['count=%s' % repr(self.count)]
        return 'read3args(%s)' % ', '.join(out)


class read3resok:
    # XDR definition:
    # struct read3resok {
    #     post_op_attr file_attributes;
    #     uint32 count;
    #     bool eof;
    #     opaque data<>;
    # };
    def __init__(self, file_attributes=None, count=None, eof=None, data=None):
        self.file_attributes = file_attributes
        self.count = count
        self.eof = eof
        self.data = data

    def __repr__(self):
        out = []
        if self.file_attributes is not None:
            out += ['file_attributes=%s' % repr(self.file_attributes)]
        if self.count is not None:
            out += ['count=%s' % repr(self.count)]
        if self.eof is not None:
            out += ['eof=%s' % repr(self.eof)]
        if self.data is not None:
            out += ['data=%s' % repr(self.data)]
        return 'read3resok(%s)' % ', '.join(out)


class read3res:
    # XDR definition:
    # union read3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         read3resok resok;
    #     default:
    #         post_op_attr resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'read3res(%s)' % ', '.join(out)


class write3args:
    # XDR definition:
    # struct write3args {
    #     nfs_fh3 file;
    #     uint64 offset;
    #     uint32 count;
    #     stable_how stable;
    #     opaque data<>;
    # };
    def __init__(self, file=None, offset=None, count=None, stable=None, data=None):
        self.file = file
        self.offset = offset
        self.count = count
        self.stable = stable
        self.data = data

    def __repr__(self):
        out = []
        if self.file is not None:
            out += ['file=%s' % repr(self.file)]
        if self.offset is not None:
            out += ['offset=%s' % repr(self.offset)]
        if self.count is not None:
            out += ['count=%s' % repr(self.count)]
        if self.stable is not None:
            out += ['stable=%s' % const.STABLE_HOW.get(self.stable, self.stable)]
        if self.data is not None:
            out += ['data=%s' % repr(self.data)]
        return 'write3args(%s)' % ', '.join(out)


class write3resok:
    # XDR definition:
    # struct write3resok {
    #     wcc_data file_wcc;
    #     uint32 count;
    #     stable_how committed;
    #     writeverf3 verf;
    # };
    def __init__(self, file_wcc=None, count=None, committed=None, verf=None):
        self.file_wcc = file_wcc
        self.count = count
        self.committed = committed
        self.verf = verf

    def __repr__(self):
        out = []
        if self.file_wcc is not None:
            out += ['file_wcc=%s' % repr(self.file_wcc)]
        if self.count is not None:
            out += ['count=%s' % repr(self.count)]
        if self.committed is not None:
            out += ['committed=%s' % const.STABLE_HOW.get(self.committed, self.committed)]
        if self.verf is not None:
            out += ['verf=%s' % repr(self.verf)]
        return 'write3resok(%s)' % ', '.join(out)


class write3res:
    # XDR definition:
    # union write3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         write3resok resok;
    #     default:
    #         wcc_data resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'write3res(%s)' % ', '.join(out)


class createhow3:
    # XDR definition:
    # union createhow3 switch(createmode3 mode) {
    #     case UNCHECKED:
    #     case GUARDED:
    #         sattr3 obj_attributes;
    #     case EXCLUSIVE:
    #         createverf3 verf;
    # };
    def __init__(self, mode=None, obj_attributes=None, verf=None):
        self.mode = mode
        self.obj_attributes = obj_attributes
        self.verf = verf

    def __repr__(self):
        out = []
        if self.mode is not None:
            out += ['mode=%s' % const.CREATEMODE3.get(self.mode, self.mode)]
        if self.obj_attributes is not None:
            out += ['obj_attributes=%s' % repr(self.obj_attributes)]
        if self.verf is not None:
            out += ['verf=%s' % repr(self.verf)]
        return 'createhow3(%s)' % ', '.join(out)


class create3args:
    # XDR definition:
    # struct create3args {
    #     diropargs3 where;
    #     createhow3 how;
    # };
    def __init__(self, where=None, how=None):
        self.where = where
        self.how = how

    def __repr__(self):
        out = []
        if self.where is not None:
            out += ['where=%s' % repr(self.where)]
        if self.how is not None:
            out += ['how=%s' % repr(self.how)]
        return 'create3args(%s)' % ', '.join(out)


class create3resok:
    # XDR definition:
    # struct CREATE3resok {
    #     post_op_fh3 obj;
    #     post_op_attr obj_attributes;
    #     wcc_data dir_wcc;
    # };
    def __init__(self, obj=None, obj_attributes=None, dir_wcc=None):
        self.obj = obj
        self.obj_attributes = obj_attributes
        self.dir_wcc = dir_wcc

    def __repr__(self):
        out = list()
        if self.obj is not None:
            out += ['obj=%s' % repr(self.obj)]
        if self.obj_attributes is not None:
            out += ['obj_attributes=%s' % repr(self.obj_attributes)]
        if self.dir_wcc is not None:
            out += ['dir_wcc=%s' % repr(self.dir_wcc)]
        return 'create3resok(%s)' % ', '.join(out)


class create3res:
    # XDR definition:
    # union CREATE3res switch (nfsstat3 status) {
    #     case NFS3_OK:
    #         CREATE3resok resok;
    #     default:
    #         CREATE3resfail resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = list()
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'create3res(%s)' % ', '.join(out)


class mkdir3args:
    # XDR definition:
    # struct mkdir3args {
    #     diropargs3 where;
    #     sattr3 attributes;
    # };
    def __init__(self, where=None, attributes=None):
        self.where = where
        self.attributes = attributes

    def __repr__(self):
        out = []
        if self.where is not None:
            out += ['where=%s' % repr(self.where)]
        if self.attributes is not None:
            out += ['attributes=%s' % repr(self.attributes)]
        return 'mkdir3args(%s)' % ', '.join(out)


class symlinkdata3:
    # XDR definition:
    # struct symlinkdata3 {
    #     sattr3 symlink_attributes;
    #     nfspath3 symlink_data;
    # };
    def __init__(self, symlink_attributes=None, symlink_data=None):
        self.symlink_attributes = symlink_attributes
        self.symlink_data = symlink_data

    def __repr__(self):
        out = []
        if self.symlink_attributes is not None:
            out += ['symlink_attributes=%s' % repr(self.symlink_attributes)]
        if self.symlink_data is not None:
            out += ['symlink_data=%s' % repr(self.symlink_data)]
        return 'symlinkdata3(%s)' % ', '.join(out)


class symlink3args:
    # XDR definition:
    # struct symlink3args {
    #     diropargs3 where;
    #     symlinkdata3 symlink;
    # };
    def __init__(self, where=None, symlink=None):
        self.where = where
        self.symlink = symlink

    def __repr__(self):
        out = []
        if self.where is not None:
            out += ['where=%s' % repr(self.where)]
        if self.symlink is not None:
            out += ['symlink=%s' % repr(self.symlink)]
        return 'symlink3args(%s)' % ', '.join(out)


class devicedata3:
    # XDR definition:
    # struct devicedata3 {
    #     sattr3 dev_attributes;
    #     specdata3 spec;
    # };
    def __init__(self, dev_attributes=None, spec=None):
        self.dev_attributes = dev_attributes
        self.spec = spec

    def __repr__(self):
        out = []
        if self.dev_attributes is not None:
            out += ['dev_attributes=%s' % repr(self.dev_attributes)]
        if self.spec is not None:
            out += ['spec=%s' % repr(self.spec)]
        return 'devicedata3(%s)' % ', '.join(out)


class mknoddata3:
    # XDR definition:
    # union mknoddata3 switch(ftype3 type) {
    #     case NF3CHR:
    #     case NF3BLK:
    #         devicedata3 device;
    #     case NF3SOCK:
    #     case NF3FIFO:
    #         sattr3 pipe_attributes;
    #     default:
    #         void;
    # };
    def __init__(self, type=None, device=None, pipe_attributes=None):
        self.type = type
        self.device = device
        self.pipe_attributes = pipe_attributes

    def __repr__(self):
        out = []
        if self.type is not None:
            out += ['type=%s' % const.FTYPE3.get(self.type, self.type)]
        if self.device is not None:
            out += ['device=%s' % repr(self.device)]
        if self.pipe_attributes is not None:
            out += ['pipe_attributes=%s' % repr(self.pipe_attributes)]
        return 'mknoddata3(%s)' % ', '.join(out)


class mknod3args:
    # XDR definition:
    # struct mknod3args {
    #     diropargs3 where;
    #     mknoddata3 what;
    # };
    def __init__(self, where=None, what=None):
        self.where = where
        self.what = what

    def __repr__(self):
        out = []
        if self.where is not None:
            out += ['where=%s' % repr(self.where)]
        if self.what is not None:
            out += ['what=%s' % repr(self.what)]
        return 'mknod3args(%s)' % ', '.join(out)


class rename3args:
    # XDR definition:
    # struct rename3args {
    #     diropargs3 from;
    #     diropargs3 to;
    # };
    def __init__(self, from_v=None, to=None):
        self.from_v = from_v
        self.to = to

    def __repr__(self):
        out = []
        if self.from_v is not None:
            out += ['from=%s' % repr(self.from_v)]
        if self.to is not None:
            out += ['to=%s' % repr(self.to)]
        return 'rename3args(%s)' % ', '.join(out)


class rename3wcc:
    # XDR definition:
    # struct rename3wcc {
    #     wcc_data fromdir_wcc;
    #     wcc_data todir_wcc;
    # };
    def __init__(self, fromdir_wcc=None, todir_wcc=None):
        self.fromdir_wcc = fromdir_wcc
        self.todir_wcc = todir_wcc

    def __repr__(self):
        out = []
        if self.fromdir_wcc is not None:
            out += ['fromdir_wcc=%s' % repr(self.fromdir_wcc)]
        if self.todir_wcc is not None:
            out += ['todir_wcc=%s' % repr(self.todir_wcc)]
        return 'rename3wcc(%s)' % ', '.join(out)


class rename3res:
    # XDR definition:
    # union rename3res switch(nfsstat3 status) {
    #     case -1:
    #         void;
    #     default:
    #         rename3wcc res;
    # };
    def __init__(self, status=None):
        self.status = status

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        return 'rename3res(%s)' % ', '.join(out)


class link3args:
    # XDR definition:
    # struct link3args {
    #     nfs_fh3 file;
    #     diropargs3 link;
    # };
    def __init__(self, file=None, link=None):
        self.file = file
        self.link = link

    def __repr__(self):
        out = []
        if self.file is not None:
            out += ['file=%s' % repr(self.file)]
        if self.link is not None:
            out += ['link=%s' % repr(self.link)]
        return 'link3args(%s)' % ', '.join(out)


class link3wcc:
    # XDR definition:
    # struct link3wcc {
    #     post_op_attr file_attributes;
    #     wcc_data linkdir_wcc;
    # };
    def __init__(self, file_attributes=None, linkdir_wcc=None):
        self.file_attributes = file_attributes
        self.linkdir_wcc = linkdir_wcc

    def __repr__(self):
        out = []
        if self.file_attributes is not None:
            out += ['file_attributes=%s' % repr(self.file_attributes)]
        if self.linkdir_wcc is not None:
            out += ['linkdir_wcc=%s' % repr(self.linkdir_wcc)]
        return 'link3wcc(%s)' % ', '.join(out)


class link3res:
    # XDR definition:
    # union link3res switch(nfsstat3 status) {
    #     case -1:
    #         void;
    #     default:
    #         link3wcc res;
    # };
    def __init__(self, status=None):
        self.status = status

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        return 'link3res(%s)' % ', '.join(out)


class readdir3args:
    # XDR definition:
    # struct readdir3args {
    #     nfs_fh3 dir;
    #     uint64 cookie;
    #     cookieverf3 cookieverf;
    #     uint32 count;
    # };
    def __init__(self, dir=None, cookie=None, cookieverf=None, count=None):
        self.dir = dir
        self.cookie = cookie
        self.cookieverf = cookieverf
        self.count = count

    def __repr__(self):
        out = []
        if self.dir is not None:
            out += ['dir=%s' % repr(self.dir)]
        if self.cookie is not None:
            out += ['cookie=%s' % repr(self.cookie)]
        if self.cookieverf is not None:
            out += ['cookieverf=%s' % repr(self.cookieverf)]
        if self.count is not None:
            out += ['count=%s' % repr(self.count)]
        return 'readdir3args(%s)' % ', '.join(out)


class entry3:
    # XDR definition:
    # struct entry3 {
    #     uint64 fileid;
    #     filename3 name;
    #     uint64 cookie;
    #     entry3 nextentry<1>;
    # };
    def __init__(self, fileid=None, name=None, cookie=None, nextentry=None):
        self.fileid = fileid
        self.name = name
        self.cookie = cookie
        self.nextentry = nextentry

    def __repr__(self):
        out = []
        if self.fileid is not None:
            out += ['fileid=%s' % repr(self.fileid)]
        if self.name is not None:
            out += ['name=%s' % repr(self.name)]
        if self.cookie is not None:
            out += ['cookie=%s' % repr(self.cookie)]
        if self.nextentry is not None:
            out += ['nextentry=%s' % repr(self.nextentry)]
        return 'entry3(%s)' % ', '.join(out)


class dirlist3:
    # XDR definition:
    # struct dirlist3 {
    #     entry3 entries<1>;
    #     bool eof;
    # };
    def __init__(self, entries=None, eof=None):
        self.entries = entries
        self.eof = eof

    def __repr__(self):
        out = []
        if self.entries is not None:
            out += ['entries=%s' % repr(self.entries)]
        if self.eof is not None:
            out += ['eof=%s' % repr(self.eof)]
        return 'dirlist3(%s)' % ', '.join(out)


class readdir3resok:
    # XDR definition:
    # struct readdir3resok {
    #     post_op_attr dir_attributes;
    #     cookieverf3 cookieverf;
    #     dirlist3 reply;
    # };
    def __init__(self, dir_attributes=None, cookieverf=None, reply=None):
        self.dir_attributes = dir_attributes
        self.cookieverf = cookieverf
        self.reply = reply

    def __repr__(self):
        out = []
        if self.dir_attributes is not None:
            out += ['dir_attributes=%s' % repr(self.dir_attributes)]
        if self.cookieverf is not None:
            out += ['cookieverf=%s' % repr(self.cookieverf)]
        if self.reply is not None:
            out += ['reply=%s' % repr(self.reply)]
        return 'readdir3resok(%s)' % ', '.join(out)


class readdir3res:
    # XDR definition:
    # union readdir3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         readdir3resok resok;
    #     default:
    #         post_op_attr resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'readdir3res(%s)' % ', '.join(out)


class readdirplus3args:
    # XDR definition:
    # struct readdirplus3args {
    #     nfs_fh3 dir;
    #     uint64 cookie;
    #     cookieverf3 cookieverf;
    #     uint32 dircount;
    #     uint32 maxcount;
    # };
    def __init__(self, dir=None, cookie=None, cookieverf=None, dircount=None, maxcount=None):
        self.dir = dir
        self.cookie = cookie
        self.cookieverf = cookieverf
        self.dircount = dircount
        self.maxcount = maxcount

    def __repr__(self):
        out = []
        if self.dir is not None:
            out += ['dir=%s' % repr(self.dir)]
        if self.cookie is not None:
            out += ['cookie=%s' % repr(self.cookie)]
        if self.cookieverf is not None:
            out += ['cookieverf=%s' % repr(self.cookieverf)]
        if self.dircount is not None:
            out += ['dircount=%s' % repr(self.dircount)]
        if self.maxcount is not None:
            out += ['maxcount=%s' % repr(self.maxcount)]
        return 'readdirplus3args(%s)' % ', '.join(out)


class entryplus3:
    # XDR definition:
    # struct entryplus3 {
    #     uint64 fileid;
    #     filename3 name;
    #     uint64 cookie;
    #     post_op_attr name_attributes;
    #     post_op_fh3 name_handle;
    #     entryplus3 nextentry<1>;
    # };
    def __init__(self, fileid=None, name=None, cookie=None, name_attributes=None, name_handle=None, nextentry=None):
        self.fileid = fileid
        self.name = name
        self.cookie = cookie
        self.name_attributes = name_attributes
        self.name_handle = name_handle
        self.nextentry = nextentry

    def __repr__(self):
        out = []
        if self.fileid is not None:
            out += ['fileid=%s' % repr(self.fileid)]
        if self.name is not None:
            out += ['name=%s' % repr(self.name)]
        if self.cookie is not None:
            out += ['cookie=%s' % repr(self.cookie)]
        if self.name_attributes is not None:
            out += ['name_attributes=%s' % repr(self.name_attributes)]
        if self.name_handle is not None:
            out += ['name_handle=%s' % repr(self.name_handle)]
        if self.nextentry is not None:
            out += ['nextentry=%s' % repr(self.nextentry)]
        return 'entryplus3(%s)' % ', '.join(out)


class dirlistplus3:
    # XDR definition:
    # struct dirlistplus3 {
    #     entryplus3 entries<1>;
    #     bool eof;
    # };
    def __init__(self, entries=None, eof=None):
        self.entries = entries
        self.eof = eof

    def __repr__(self):
        out = []
        if self.entries is not None:
            out += ['entries=%s' % repr(self.entries)]
        if self.eof is not None:
            out += ['eof=%s' % repr(self.eof)]
        return 'dirlistplus3(%s)' % ', '.join(out)


class readdirplus3resok:
    # XDR definition:
    # struct readdirplus3resok {
    #     post_op_attr dir_attributes;
    #     cookieverf3 cookieverf;
    #     dirlistplus3 reply;
    # };
    def __init__(self, dir_attributes=None, cookieverf=None, reply=None):
        self.dir_attributes = dir_attributes
        self.cookieverf = cookieverf
        self.reply = reply

    def __repr__(self):
        out = []
        if self.dir_attributes is not None:
            out += ['dir_attributes=%s' % repr(self.dir_attributes)]
        if self.cookieverf is not None:
            out += ['cookieverf=%s' % repr(self.cookieverf)]
        if self.reply is not None:
            out += ['reply=%s' % repr(self.reply)]
        return 'readdirplus3resok(%s)' % ', '.join(out)


class readdirplus3res:
    # XDR definition:
    # union readdirplus3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         readdirplus3resok resok;
    #     default:
    #         post_op_attr resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'readdirplus3res(%s)' % ', '.join(out)


class fsstat3resok:
    # XDR definition:
    # struct fsstat3resok {
    #     post_op_attr obj_attributes;
    #     uint64 tbytes;
    #     uint64 fbytes;
    #     uint64 abytes;
    #     uint64 tfiles;
    #     uint64 ffiles;
    #     uint64 afiles;
    #     uint32 invarsec;
    # };
    def __init__(self, obj_attributes=None, tbytes=None, fbytes=None, abytes=None, tfiles=None, ffiles=None,
                 afiles=None, invarsec=None):
        self.obj_attributes = obj_attributes
        self.tbytes = tbytes
        self.fbytes = fbytes
        self.abytes = abytes
        self.tfiles = tfiles
        self.ffiles = ffiles
        self.afiles = afiles
        self.invarsec = invarsec

    def __repr__(self):
        out = []
        if self.obj_attributes is not None:
            out += ['obj_attributes=%s' % repr(self.obj_attributes)]
        if self.tbytes is not None:
            out += ['tbytes=%s' % repr(self.tbytes)]
        if self.fbytes is not None:
            out += ['fbytes=%s' % repr(self.fbytes)]
        if self.abytes is not None:
            out += ['abytes=%s' % repr(self.abytes)]
        if self.tfiles is not None:
            out += ['tfiles=%s' % repr(self.tfiles)]
        if self.ffiles is not None:
            out += ['ffiles=%s' % repr(self.ffiles)]
        if self.afiles is not None:
            out += ['afiles=%s' % repr(self.afiles)]
        if self.invarsec is not None:
            out += ['invarsec=%s' % repr(self.invarsec)]
        return 'fsstat3resok(%s)' % ', '.join(out)


class fsstat3res:
    # XDR definition:
    # union fsstat3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         fsstat3resok resok;
    #     default:
    #         post_op_attr resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'fsstat3res(%s)' % ', '.join(out)


class fsinfo3resok:
    # XDR definition:
    # struct fsinfo3resok {
    #     post_op_attr obj_attributes;
    #     uint32 rtmax;
    #     uint32 rtpref;
    #     uint32 rtmult;
    #     uint32 wtmax;
    #     uint32 wtpref;
    #     uint32 wtmult;
    #     uint32 dtpref;
    #     uint64 maxfilesize;
    #     nfstime3 time_delta;
    #     uint32 properties;
    # };
    def __init__(self, obj_attributes=None, rtmax=None, rtpref=None, rtmult=None, wtmax=None, wtpref=None,
                 wtmult=None, dtpref=None, maxfilesize=None, time_delta=None, properties=None):
        self.obj_attributes = obj_attributes
        self.rtmax = rtmax
        self.rtpref = rtpref
        self.rtmult = rtmult
        self.wtmax = wtmax
        self.wtpref = wtpref
        self.wtmult = wtmult
        self.dtpref = dtpref
        self.maxfilesize = maxfilesize
        self.time_delta = time_delta
        self.properties = properties

    def __repr__(self):
        out = []
        if self.obj_attributes is not None:
            out += ['obj_attributes=%s' % repr(self.obj_attributes)]
        if self.rtmax is not None:
            out += ['rtmax=%s' % repr(self.rtmax)]
        if self.rtpref is not None:
            out += ['rtpref=%s' % repr(self.rtpref)]
        if self.rtmult is not None:
            out += ['rtmult=%s' % repr(self.rtmult)]
        if self.wtmax is not None:
            out += ['wtmax=%s' % repr(self.wtmax)]
        if self.wtpref is not None:
            out += ['wtpref=%s' % repr(self.wtpref)]
        if self.wtmult is not None:
            out += ['wtmult=%s' % repr(self.wtmult)]
        if self.dtpref is not None:
            out += ['dtpref=%s' % repr(self.dtpref)]
        if self.maxfilesize is not None:
            out += ['maxfilesize=%s' % repr(self.maxfilesize)]
        if self.time_delta is not None:
            out += ['time_delta=%s' % repr(self.time_delta)]
        if self.properties is not None:
            out += ['properties=%s' % repr(self.properties)]
        return 'fsinfo3resok(%s)' % ', '.join(out)


class fsinfo3res:
    # XDR definition:
    # union fsinfo3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         fsinfo3resok resok;
    #     default:
    #         post_op_attr resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'fsinfo3res(%s)' % ', '.join(out)


class pathconf3resok:
    # XDR definition:
    # struct pathconf3resok {
    #     post_op_attr obj_attributes;
    #     uint32 linkmax;
    #     uint32 name_max;
    #     bool no_trunc;
    #     bool chown_restricted;
    #     bool case_insensitive;
    #     bool case_preserving;
    # };
    def __init__(self, obj_attributes=None, linkmax=None, name_max=None, no_trunc=None, chown_restricted=None,
                 case_insensitive=None, case_preserving=None):
        self.obj_attributes = obj_attributes
        self.linkmax = linkmax
        self.name_max = name_max
        self.no_trunc = no_trunc
        self.chown_restricted = chown_restricted
        self.case_insensitive = case_insensitive
        self.case_preserving = case_preserving

    def __repr__(self):
        out = []
        if self.obj_attributes is not None:
            out += ['obj_attributes=%s' % repr(self.obj_attributes)]
        if self.linkmax is not None:
            out += ['linkmax=%s' % repr(self.linkmax)]
        if self.name_max is not None:
            out += ['name_max=%s' % repr(self.name_max)]
        if self.no_trunc is not None:
            out += ['no_trunc=%s' % repr(self.no_trunc)]
        if self.chown_restricted is not None:
            out += ['chown_restricted=%s' % repr(self.chown_restricted)]
        if self.case_insensitive is not None:
            out += ['case_insensitive=%s' % repr(self.case_insensitive)]
        if self.case_preserving is not None:
            out += ['case_preserving=%s' % repr(self.case_preserving)]
        return 'pathconf3resok(%s)' % ', '.join(out)


class pathconf3res:
    # XDR definition:
    # union pathconf3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         pathconf3resok resok;
    #     default:
    #         post_op_attr resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'pathconf3res(%s)' % ', '.join(out)


class commit3args:
    # XDR definition:
    # struct commit3args {
    #     nfs_fh3 file;
    #     uint64 offset;
    #     uint32 count;
    # };
    def __init__(self, file=None, offset=None, count=None):
        self.file = file
        self.offset = offset
        self.count = count

    def __repr__(self):
        out = []
        if self.file is not None:
            out += ['file=%s' % repr(self.file)]
        if self.offset is not None:
            out += ['offset=%s' % repr(self.offset)]
        if self.count is not None:
            out += ['count=%s' % repr(self.count)]
        return 'commit3args(%s)' % ', '.join(out)


class commit3resok:
    # XDR definition:
    # struct commit3resok {
    #     wcc_data file_wcc;
    #     writeverf3 verf;
    # };
    def __init__(self, file_wcc=None, verf=None):
        self.file_wcc = file_wcc
        self.verf = verf

    def __repr__(self):
        out = []
        if self.file_wcc is not None:
            out += ['file_wcc=%s' % repr(self.file_wcc)]
        if self.verf is not None:
            out += ['verf=%s' % repr(self.verf)]
        return 'commit3resok(%s)' % ', '.join(out)


class commit3res:
    # XDR definition:
    # union commit3res switch(nfsstat3 status) {
    #     case NFS3_OK:
    #         commit3resok resok;
    #     default:
    #         wcc_data resfail;
    # };
    def __init__(self, status=None, resok=None):
        self.status = status
        self.resok = resok

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['status=%s' % const.NFSSTAT3.get(self.status, self.status)]
        if self.resok is not None:
            out += ['resok=%s' % repr(self.resok)]
        return 'commit3res(%s)' % ', '.join(out)


class setaclargs:
    # XDR definition:
    # struct setaclargs {
    #     diropargs3 dargs;
    #     write3args wargs;
    # };
    def __init__(self, dargs=None, wargs=None):
        self.dargs = dargs
        self.wargs = wargs

    def __repr__(self):
        out = []
        if self.dargs is not None:
            out += ['dargs=%s' % repr(self.dargs)]
        if self.wargs is not None:
            out += ['wargs=%s' % repr(self.wargs)]
        return 'setaclargs(%s)' % ', '.join(out)


class mountres3_ok:
    # XDR definition:
    # struct mountres3_ok {
    #     fhandle3 fhandle;
    #     int auth_flavors<>;
    # };
    def __init__(self, fhandle=None, auth_flavors=None):
        self.fhandle = fhandle
        self.auth_flavors = auth_flavors

    def __repr__(self):
        out = []
        if self.fhandle is not None:
            out += ['fhandle=%s' % repr(self.fhandle)]
        if self.auth_flavors is not None:
            out += ['auth_flavors=%s' % repr(self.auth_flavors)]
        return 'mountres3_ok(%s)' % ', '.join(out)


class mountres3:
    # XDR definition:
    # union mountres3 switch(mountstat3 fhs_status) {
    #     case MNT3_OK:
    #         mountres3_ok mountinfo;
    #     default:
    #         void;
    # };
    def __init__(self, fhs_status=None, mountinfo=None):
        self.status = fhs_status
        self.mountinfo = mountinfo

    def __repr__(self):
        out = []
        if self.status is not None:
            out += ['fhs_status=%s' % const.MOUNTSTAT3.get(self.status, self.status)]
        if self.mountinfo is not None:
            out += ['mountinfo=%s' % repr(self.mountinfo)]
        return 'mountres3(%s)' % ', '.join(out)


class mountbody:
    # XDR definition:
    # struct mountbody {
    #     name ml_hostname;
    #     dirpath ml_directory;
    #     mountlist ml_next;
    # };
    def __init__(self, ml_hostname=None, ml_directory=None, ml_next=None):
        self.ml_hostname = ml_hostname
        self.ml_directory = ml_directory
        self.ml_next = ml_next

    def __repr__(self):
        out = []
        if self.ml_hostname is not None:
            out += ['ml_hostname=%s' % repr(self.ml_hostname)]
        if self.ml_directory is not None:
            out += ['ml_directory=%s' % repr(self.ml_directory)]
        if self.ml_next is not None:
            out += ['ml_next=%s' % repr(self.ml_next)]
        return 'mountbody(%s)' % ', '.join(out)


class groupnode:
    # XDR definition:
    # struct groupnode {
    #     name gr_name;
    #     groups gr_next;
    # };
    def __init__(self, gr_name=None, gr_next=None):
        self.gr_name = gr_name
        self.gr_next = gr_next

    def __repr__(self):
        out = []
        if self.gr_name is not None:
            out += ['gr_name=%s' % repr(self.gr_name)]
        if self.gr_next is not None:
            out += ['gr_next=%s' % repr(self.gr_next)]
        return 'groupnode(%s)' % ', '.join(out)


class exportnode:
    # XDR definition:
    # struct exportnode {
    #     dirpath ex_dir;
    #     groups ex_groups;
    #     exports ex_next;
    # };
    def __init__(self, ex_dir=None, ex_groups=None, ex_next=None):
        self.ex_dir = ex_dir
        self.ex_groups = ex_groups
        self.ex_next = ex_next

    def __repr__(self):
        out = []
        if self.ex_dir is not None:
            out += ['ex_dir=%s' % repr(self.ex_dir)]
        if self.ex_groups is not None:
            out += ['ex_groups=%s' % repr(self.ex_groups)]
        if self.ex_next is not None:
            out += ['ex_next=%s' % repr(self.ex_next)]
        return 'exportnode(%s)' % ', '.join(out)



# https://raw.githubusercontent.com/CharmingYang0/NfsClient/master/pyNfsClient/nfs3.py

import logging
from functools import wraps
from anfs.protocol.rpc import RPC

from anfs.protocol.nfs3.pack import nfs_pro_v3Packer, nfs_pro_v3Unpacker, nextiter, NFSFileEntry
from anfs.protocol.nfs3.rtypes import (nfs_fh3, set_uint32, set_uint64, sattr3, set_time, diropargs3, setattr3args, create3args,
					 mkdir3args, symlink3args, commit3args, sattrguard3, access3args, readdir3args, readdirplus3args,
					 read3args, write3args, createhow3, symlinkdata3, mknod3args, mknoddata3, devicedata3, specdata3,
					 link3args, rename3args, nfstime3)
from anfs.protocol.nfs3.messages import (NFS3_PROCEDURE_NULL, NFS3_PROCEDURE_GETATTR, NFS3_PROCEDURE_SETATTR, NFS3_PROCEDURE_LOOKUP,
					NFS3_PROCEDURE_ACCESS, NFS3_PROCEDURE_READLINK, NFS3_PROCEDURE_READ, NFS3_PROCEDURE_WRITE,
					NFS3_PROCEDURE_CREATE, NFS3_PROCEDURE_MKDIR, NFS3_PROCEDURE_SYMLINK, NFS3_PROCEDURE_MKNOD,
					NFS3_PROCEDURE_REMOVE, NFS3_PROCEDURE_RMDIR, NFS3_PROCEDURE_RENAME, NFS3_PROCEDURE_LINK,
					NFS3_PROCEDURE_READDIR, NFS3_PROCEDURE_READDIRPLUS, NFS3_PROCEDURE_FSSTAT, NFS3_PROCEDURE_FSINFO,
					NFS3_PROCEDURE_PATHCONF, NFS3_PROCEDURE_COMMIT, NFS_PROGRAM, NFS_V3, NF3BLK, NF3CHR, NF3FIFO,
					NF3SOCK, time_how, DONT_CHANGE, SET_TO_CLIENT_TIME, SET_TO_SERVER_TIME)

from anfs.protocol.nfs3 import logger

class NFSAccessError(Exception):
	pass


class NFSv3Client:
	def __init__(self, root_handle, target, credential=None, encoding='utf-8'):
		self.root_handle = root_handle
		self.target = target
		self.credential = credential
		self.rpc:RPC = None
		self.host = ''
		self.encoding = encoding
		self.__handle_id = 1
		self.__handles = {
			0: root_handle
		}
		# NFSv3 doesn't support getting full path from handle, so we need to keep a map of handles to names
		self.__handle_name_map = {
			root_handle: '/'
		}
		self.__parent_handles = {}
		self.__handle_reverse_map = {
			root_handle : 0
		}

	async def __aenter__(self):
		_, err = await self.connect()
		if err is not None:
			raise err
		return self
	
	async def __aexit__(self, exc_type, exc, tb):
		await self.disconnect()

	def handle_to_path(self, handle, parts=None):
		if isinstance(handle, int):
			handle = self.__handles[handle]
		if handle not in self.__handle_name_map:
			return ''
		if parts is None:
			parts = []
		if self.__handle_name_map[handle] in parts:
			# We have a loop!!! This is bad
			return '/'.join(parts)

		parts.insert(0, self.__handle_name_map[handle])
		if handle == self.root_handle:
			return '/'.join(parts)[1:]
		return self.handle_to_path(self.__parent_handles[handle], parts)

	def register_handle(self, handle, name, parent_handle):
		if handle is None:
			return None
		if handle in self.__handle_reverse_map:
			return self.__handle_reverse_map[handle]
		self.__handles[self.__handle_id] = handle
		self.__handle_reverse_map[handle] = self.__handle_id
		self.__handle_id += 1
		rethandle = self.__handle_id - 1
		if name == '.' or name == '..':
			return rethandle
		self.__handle_name_map[handle] = name
		self.__parent_handles[handle] = parent_handle
		return rethandle
	
	async def connect(self):
		try:
			self.rpc = RPC(self.target, self.credential)
			return await self.rpc.connect(NFS_PROGRAM, NFS_V3)
		except Exception as e:
			return False, e
		
	async def disconnect(self):
		if self.rpc is not None:
			return await self.rpc.disconnect()

	async def nfs_request(self, procedure, args):
		return await self.rpc.call(NFS_PROGRAM, NFS_V3, procedure, args)
		

	async def null(self):
		try:
			_, err = await self.rpc.call(NFS_PROGRAM, NFS_V3, NFS3_PROCEDURE_NULL, b'')
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			return False, e

	@classmethod
	async def get_sattr3(cls, mode=None, uid=None, gid=None, size=None, atime_flag=None, atime_s=0, atime_ns=0,
				   mtime_flag=None, mtime_s=0, mtime_ns=0):
		if atime_flag not in time_how:
			raise ValueError("atime flag must be one of %s" % time_how.keys())

		if mtime_flag not in time_how:
			raise ValueError("mtime flag must be one of %s" % time_how.keys())

		attrs = sattr3(mode=set_uint32(True, int(mode)) if mode is not None else set_uint32(False),
					   uid=set_uint32(True, int(uid)) if uid is not None else set_uint32(False),
					   gid=set_uint32(True, int(gid)) if gid is not None else set_uint32(False),
					   size=set_uint64(True, int(size)) if size is not None else set_uint64(False),
					   atime=set_time(SET_TO_CLIENT_TIME, nfstime3(int(atime_s), int(atime_ns)))
							 if atime_flag == SET_TO_CLIENT_TIME else set_time(atime_flag),
					   mtime=set_time(SET_TO_CLIENT_TIME, nfstime3(int(mtime_s), int(mtime_ns)))
							 if mtime_flag == SET_TO_CLIENT_TIME else set_time(mtime_flag))
		return attrs

	async def getattr(self, file_handle):
		try:
			ofh = file_handle
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_fhandle3(file_handle)

			logger.debug("NFSv3 procedure %d: GETATTR on %s" % (NFS3_PROCEDURE_GETATTR, self.host))
			data, err = await self.nfs_request(NFS3_PROCEDURE_GETATTR, packer.get_buffer())
			if err is not None:
				raise err
			
			unpacker = nfs_pro_v3Unpacker(data)
			response = unpacker.unpack_getattr3res()
			
			if response['status'] != 0:
				raise NFSAccessError(response.status)

			fe = NFSFileEntry.from_attrs(response['attributes'])
			fe.handle = ofh
			
			return fe, None
		except Exception as e:
			return False, e

	# TODO: Implement setattr
	async def setattr(self, file_handle, mode=None, uid=None, gid=None, size=None,
				atime_flag=SET_TO_SERVER_TIME, atime_s=None, atime_us=None,
				mtime_flag=SET_TO_SERVER_TIME, mtime_s=None, mtime_us=None,
				check=False, obj_ctime=None):
		try:
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			attrs = await self.get_sattr3(mode, uid, gid, size, atime_flag, atime_s, atime_us, mtime_flag, mtime_s, mtime_us)
			packer.pack_setattr3args(setattr3args(object=nfs_fh3(file_handle),
												new_attributes=attrs,
												guard=sattrguard3(check=check, ctime=obj_ctime)))

			logger.debug("NFSv3 procedure %d: GETATTR on %s" % (NFS3_PROCEDURE_SETATTR, self.host))
			data, err = await self.nfs_request(NFS3_PROCEDURE_SETATTR, packer.get_buffer())
			if err is not None:
				raise err

			unpacker = nfs_pro_v3Unpacker(data)
			return unpacker.unpack_setattr3res(), None
		except Exception as e:
			return False, e


	async def lookup(self, dir_handle, file_folder):
		try:
			dir_handle = self.__handles[dir_handle]

			packer = nfs_pro_v3Packer()
			packer.pack_diropargs3(diropargs3(dir=nfs_fh3(dir_handle), name=file_folder.encode()))

			logger.debug("NFSv3 procedure %d: LOOKUP on %s" % (NFS3_PROCEDURE_LOOKUP, self.host))
			data, err = await self.nfs_request(NFS3_PROCEDURE_LOOKUP, packer.get_buffer())
			if err is not None:
				raise err
			
			unpacker = nfs_pro_v3Unpacker(data)
			response = unpacker.unpack_lookup3res(data_format='json')
			
			if response['status'] != 0:
				return False, None
			
			entry = NFSFileEntry.from_attrs(response['resok']['obj_attributes']['attributes'])
			entry.name = file_folder
			entry.handle = self.register_handle(response['resok']['object']['data'], file_folder, dir_handle)
			return entry, None
		
		except Exception as e:
			return False, e

	async def access(self, file_handle, access_option):
		try:
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_access3args(access3args(object=nfs_fh3(file_handle), access=access_option))

			logger.debug("NFSv3 procedure %d: ACCESS on %s" % (NFS3_PROCEDURE_ACCESS, self.host))
			data, err = await self.nfs_request(NFS3_PROCEDURE_ACCESS, packer.get_buffer())
			if err is not None:
				raise err
			unpacker = nfs_pro_v3Unpacker(data)
			response = unpacker.unpack_access3res()

			if response['status'] != 0:
				raise NFSAccessError(response.status)

			return response['resok']['access'], None
		except Exception as e:
			return False, e

	async def readlink(self, file_handle):
		try:
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_fhandle3(file_handle)

			logger.debug("NFSv3 procedure %d: READLINK on %s" % (NFS3_PROCEDURE_READLINK, self.host))
			data, err = await self.nfs_request(NFS3_PROCEDURE_READLINK, packer.get_buffer())
			if err is not None:
				raise err
			
			unpacker = nfs_pro_v3Unpacker(data)
			result = unpacker.unpack_readlink3res()
			if result['status'] != 0:
				raise NFSAccessError(result.status)
			return result['resok']['data'].decode(self.encoding), None
		except Exception as e:
			return False, e

	async def read(self, file_handle, offset=0, chunk_count=1024 * 1024):
		try:
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_read3args(read3args(file=nfs_fh3(file_handle), offset=offset, count=chunk_count))

			logger.debug("NFSv3 procedure %d: READ on %s" % (NFS3_PROCEDURE_READ, self.host))
			data, err = await self.nfs_request(NFS3_PROCEDURE_READ, packer.get_buffer())
			if err is not None:
				raise err
			
			unpacker = nfs_pro_v3Unpacker(data)
			response = unpacker.unpack_read3res()
			if response['status'] != 0:
				raise NFSAccessError(response.status)
			
			return response['resok']['data'], None

		except Exception as e:
			return False, e

	async def write(self, file_handle, offset, count, content, stable_how, auth=None):
		try:
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_write3args(write3args(file=nfs_fh3(file_handle),
											offset=offset,
											count=count,
											stable=stable_how,
											data=content))

			logger.debug("NFSv3 procedure %d: WRITE on %s" % (NFS3_PROCEDURE_WRITE, self.host))
			res, err = await self.nfs_request(NFS3_PROCEDURE_WRITE, packer.get_buffer())
			if err is not None:
				raise err
			
			unpacker = nfs_pro_v3Unpacker(res)
			return unpacker.unpack_write3res(), None
		except Exception as e:
			return False, e

	async def create(self, dir_handle, file_name, create_mode, mode=None, uid=None, gid=None, size=None,
			   atime_flag=SET_TO_SERVER_TIME, atime_s=None, atime_us=None,
				mtime_flag=SET_TO_SERVER_TIME, mtime_s=None, mtime_us=None,
			   verf='0', auth=None):
		try:
			dir_handle = self.__handles[dir_handle]	
			packer = nfs_pro_v3Packer()
			attrs = await self.get_sattr3(mode, uid, gid, size, atime_flag, atime_s, atime_us, mtime_flag, mtime_s, mtime_us)
			packer.pack_create3args(create3args(where=diropargs3(dir=nfs_fh3(dir_handle), name=file_name.encode()),
												how=createhow3(mode=create_mode, obj_attributes=attrs, verf=verf)))

			logger.debug("NFSv3 procedure %d: CREATE on %s" % (NFS3_PROCEDURE_CREATE, self.host))
			data, err = await self.nfs_request(NFS3_PROCEDURE_CREATE, packer.get_buffer())
			if err is not None:
				raise err
			
			unpacker = nfs_pro_v3Unpacker(data)
			response = unpacker.unpack_create3res()
			if response['status'] != 0:
				raise NFSAccessError(response.status)
			
			entry = NFSFileEntry.from_attrs(response['resok']['obj_attributes']['attributes'])
			entry.name = file_name
			entry.handle = self.register_handle(response['resok']['obj']['handle']['data'], file_name, dir_handle)
			return entry, None
		except Exception as e:
			return False, e

	async def mkdir(self, dir_handle, dir_name, mode=None, uid=None, gid=None,
			  atime_flag=SET_TO_SERVER_TIME, atime_s=None, atime_us=None,
			  mtime_flag=SET_TO_SERVER_TIME, mtime_s=None, mtime_us=None):

		try:
			dir_handle = self.__handles[dir_handle]
			packer = nfs_pro_v3Packer()
			attrs = await self.get_sattr3(mode, uid, gid, None, atime_flag, atime_s, atime_us, mtime_flag, mtime_s, mtime_us)
			packer.pack_mkdir3args(mkdir3args(where=diropargs3(dir=nfs_fh3(dir_handle), name=dir_name.encode()),
											attributes=attrs))

			logger.debug("NFSv3 procedure %d: MKDIR on %s" % (NFS3_PROCEDURE_MKDIR, self.host))
			res, err = await self.nfs_request(NFS3_PROCEDURE_MKDIR, packer.get_buffer())
			if err is not None:
				raise err
			unpacker = nfs_pro_v3Unpacker(res)
			response = unpacker.unpack_mkdir3res()
			if response['status'] != 0:
				raise NFSAccessError(response.status)
			
			entry = NFSFileEntry.from_attrs(response['resok']['obj_attributes']['attributes'])
			entry.name = dir_name
			entry.handle = self.register_handle(response['resok']['obj']['handle']['data'], dir_name, dir_handle)
			return entry, None
		except Exception as e:
			import traceback
			traceback.print_exc()
			return False, e

	async def symlink(self, dir_handle, link_name, link_to_path, auth=None):
		try:
			dir_handle = self.__handles[dir_handle]
			packer = nfs_pro_v3Packer()
			attrs = await self.get_sattr3(mode=None, size=None, uid=None, gid=None, atime_flag=DONT_CHANGE, mtime_flag=DONT_CHANGE)
			packer.pack_symlink3args(symlink3args(where=diropargs3(dir=nfs_fh3(dir_handle),
																name=link_name.encode()),
												symlink=symlinkdata3(symlink_attributes=attrs,
																	symlink_data=link_to_path.encode())))

			logger.debug("NFSv3 procedure %d: SYMLINK on %s" % (NFS3_PROCEDURE_SYMLINK, self.host))
			res, err = await self.nfs_request(NFS3_PROCEDURE_SYMLINK, packer.get_buffer())
			if err is not None:
				raise err

			unpacker = nfs_pro_v3Unpacker(res)
			result = unpacker.unpack_symlink3res()
			if result['status'] != 0:
				raise NFSAccessError(result.status)
			return True, None
		
		except Exception as e:
			return False, e

	async def mknod(self, dir_handle, file_name, ftype,
			  mode=None, uid=None, gid=None,
			  atime_flag=SET_TO_SERVER_TIME, atime_s=None, atime_us=None,
			  mtime_flag=SET_TO_SERVER_TIME, mtime_s=None, mtime_us=None,
			  spec_major=0, spec_minor=0, auth=None):
		
		try:
			dir_handle = self.__handles[dir_handle]
			packer = nfs_pro_v3Packer()
			attrs = await self.get_sattr3(mode, uid, gid, None, atime_flag, atime_s, atime_us, mtime_flag, mtime_s, mtime_us)
			if ftype in (NF3CHR, NF3BLK):
				spec = specdata3(major=spec_major, minor=spec_minor)
				what = mknoddata3(type=ftype, device=devicedata3(dev_attributes=attrs, spec=spec))
			elif ftype in (NF3SOCK, NF3FIFO):
				what = mknoddata3(type=ftype, pipe_attributes=attrs)
			else:
				raise ValueError("ftype must be one of [%d, %d, %d, %d]" % (NF3CHR, NF3BLK, NF3SOCK, NF3FIFO))
			packer.pack_mknod3args(mknod3args(where=diropargs3(dir=nfs_fh3(dir_handle),
															name=file_name.encode()),
											what=what))

			logger.debug("NFSv3 procedure %d: MKNOD on %s" % (NFS3_PROCEDURE_MKNOD, self.host))
			res = self.nfs_request(NFS3_PROCEDURE_MKNOD, packer.get_buffer())

			unpacker = nfs_pro_v3Unpacker(res)
			return unpacker.unpack_mknod3res(), None
		except Exception as e:
			return False, e


	async def remove(self, dir_handle, file_name):
		try:
			dir_handle = self.__handles[dir_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_diropargs3(diropargs3(dir=nfs_fh3(dir_handle), name=file_name.encode()))

			logger.debug("NFSv3 procedure %d: REMOVE on %s" % (NFS3_PROCEDURE_REMOVE, self.host))
			res, err = await self.nfs_request(NFS3_PROCEDURE_REMOVE, packer.get_buffer())
			if err is not None:
				raise err

			unpacker = nfs_pro_v3Unpacker(res)
			response = unpacker.unpack_remove3res()
			if response['status'] != 0:
				raise NFSAccessError(response.status)
			return True, None
		except Exception as e:
			return False, e

	async def rmdir(self, dir_handle, dir_name):
		try:
			dir_handle = self.__handles[dir_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_diropargs3(diropargs3(dir=nfs_fh3(dir_handle), name=dir_name.encode()))

			logger.debug("NFSv3 procedure %d: RMDIR on %s" % (NFS3_PROCEDURE_RMDIR, self.host))
			res, err = await self.nfs_request(NFS3_PROCEDURE_RMDIR, packer.get_buffer())
			if err is not None:
				raise err

			unpacker = nfs_pro_v3Unpacker(res)
			response = unpacker.unpack_rmdir3res()
			if response['status'] != 0:
				raise NFSAccessError(response.status)

			return True, None
		except Exception as e:
			return False, e

	async def rename(self, dir_handle_from, from_name, dir_handle_to, to_name, auth=None):
		try:
			dir_handle_from = self.__handles[dir_handle_from]
			dir_handle_to = self.__handles[dir_handle_to]

			packer = nfs_pro_v3Packer()
			packer.pack_rename3args(rename3args(from_v=diropargs3(dir=nfs_fh3(dir_handle_from),
																name=from_name.encode()),
												to=diropargs3(dir=nfs_fh3(dir_handle_to),
															name=to_name.encode())))

			logger.debug("NFSv3 procedure %d: RENAME on %s" % (NFS3_PROCEDURE_RENAME, self.host))
			res, err = await self.nfs_request(NFS3_PROCEDURE_RENAME, packer.get_buffer())

			unpacker = nfs_pro_v3Unpacker(res)
			return unpacker.unpack_rename3res(), None
		except Exception as e:
			return False, e

	
	async def link(self, file_handle, link_to_dir_handle, link_name, auth=None):
		try:
			file_handle = self.__handles[file_handle]
			link_to_dir_handle = self.__handles[link_to_dir_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_link3args(link3args(file=nfs_fh3(file_handle),
											link=diropargs3(dir=nfs_fh3(link_to_dir_handle), name=link_name.encode())))

			logger.debug("NFSv3 procedure %d: LINK on %s" % (NFS3_PROCEDURE_LINK, self.host))
			res, err = await self.nfs_request(NFS3_PROCEDURE_LINK, packer.get_buffer())
			if err is not None:
				raise err
			
			unpacker = nfs_pro_v3Unpacker(res)
			return unpacker.unpack_link3res(), None
		except Exception as e:
			return False, e

	async def readdir(self, dir_handle, cookie=0, cookie_verf='0', count=4096):
		try:
			dir_handle = self.__handles[dir_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_readdir3args(readdir3args(dir=nfs_fh3(dir_handle),
												cookie=cookie,
												cookieverf=cookie_verf.encode(),
												count=count))

			logger.debug("NFSv3 procedure %d: READDIR on %s" % (NFS3_PROCEDURE_READDIR, self.host))
			res, err =  await self.nfs_request(NFS3_PROCEDURE_READDIR, packer.get_buffer())
			if err is not None:
				raise err
			
			unpacker = nfs_pro_v3Unpacker(res)
			fl = unpacker.unpack_readdir3res(data_format='')
			if fl.status != 0:
				raise NFSAccessError(fl.status)
			
			last_cookie = 0
			for entry in nextiter(fl.resok.reply.entries):
				last_cookie = entry['cookie']
				entry, realhandle = NFSFileEntry.from_dict(entry, encoding=self.encoding)
				entry.handle = self.register_handle(realhandle, entry.name, dir_handle)

				yield entry, None
			
			if fl.resok.reply.eof is False:
				async for entry, err in self.readdirplus(dir_handle, last_cookie, cookie_verf, count):
					entry, realhandle = NFSFileEntry.from_dict(entry, encoding=self.encoding)
					entry.handle = self.register_handle(realhandle, entry.name, dir_handle)
					yield entry, err

		except Exception as e:
			yield False, e
	
	# This version was provided by PTG
	# The original version was not working properly
	async def readdirplus(self, dir_handle, cookie=0, cookie_verf='0', dircount=4096, maxcount=32768):
		try:
			real_dir_handle = self.__handles[dir_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_readdirplus3args(readdirplus3args(dir=nfs_fh3(real_dir_handle),
				cookie=cookie,
				cookieverf=cookie_verf.encode(),
				dircount=dircount,
				maxcount=maxcount)
			)
			
			logger.debug("NFSv3 procedure %d: READDIRPLUS on %s" % (NFS3_PROCEDURE_READDIRPLUS, self.host))
			res, err = await self.nfs_request(NFS3_PROCEDURE_READDIRPLUS, packer.get_buffer())
			if err is not None:
				raise err
			unpacker = nfs_pro_v3Unpacker(res)
			fl = unpacker.unpack_readdirplus3res(data_format='')
			if fl.status != 0:
				raise NFSAccessError(fl.status)
			
			last_cookie = 0
			for entry in nextiter(fl.resok.reply.entries):
				last_cookie = entry['cookie']
				entry, realhandle = NFSFileEntry.from_dict(entry, encoding=self.encoding)
				entry.handle = self.register_handle(realhandle, entry.name, real_dir_handle)
				yield entry, None
						   
			if fl.resok.reply.eof is False:
				async for entry, err in self.readdirplus(dir_handle, last_cookie, cookie_verf, dircount, maxcount):
					yield entry, err

		except Exception as e:
			yield False, e

	#async def readdirplus(self, dir_handle, cookie=0, cookie_verf='0', dircount=4096, maxcount=32768):
	#	try:
	#		dir_handle = self.__handles[dir_handle]
	#		packer = nfs_pro_v3Packer()
	#		packer.pack_readdirplus3args(readdirplus3args(dir=nfs_fh3(dir_handle),
	#													cookie=cookie,
	#													cookieverf=cookie_verf.encode(),
	#													dircount=dircount,
	#													maxcount=maxcount))
	#
	#		logger.debug("NFSv3 procedure %d: READDIRPLUS on %s" % (NFS3_PROCEDURE_READDIRPLUS, self.host))
	#		res, err = await self.nfs_request(NFS3_PROCEDURE_READDIRPLUS, packer.get_buffer())
	#		if err is not None:
	#			raise err
	#
	#		
	#		unpacker = nfs_pro_v3Unpacker(res)
	#		fl = unpacker.unpack_readdirplus3res(data_format='')
	#		if fl.status != 0:
	#			raise NFSAccessError(fl.status)
	#		
	#		last_cookie = 0
	#		for entry in nextiter(fl.resok.reply.entries):
	#			last_cookie = entry['cookie']
	#			entry, realhandle = NFSFileEntry.from_dict(entry, encoding=self.encoding)
	#			entry.handle = self.register_handle(realhandle, entry.name, dir_handle)
	#
	#			yield entry, None
	#		
	#		if fl.resok.reply.eof is False:
	#			async for entry, err in self.readdirplus(dir_handle, last_cookie, cookie_verf, dircount, maxcount):
	#				entry, realhandle = NFSFileEntry.from_dict(entry, encoding=self.encoding)
	#				entry.handle = self.register_handle(realhandle, entry.name, dir_handle)
	#				yield entry, err
	#
	#	except Exception as e:
	#		yield False, e
	#
	
	async def fsstat(self, file_handle, auth=None):
		try:
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_fhandle3(file_handle)

			logger.debug("NFSv3 procedure %d: FSSTAT on %s" % (NFS3_PROCEDURE_FSSTAT, self.host))
			data = self.nfs_request(NFS3_PROCEDURE_FSSTAT, packer.get_buffer())

			unpacker = nfs_pro_v3Unpacker(data)
			return unpacker.unpack_fsstat3res(), None
		except Exception as e:
			return False, e

	
	async def fsinfo(self, file_handle, auth=None):
		try:
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_fhandle3(file_handle)

			logger.debug("NFSv3 procedure %d: FSINFO on %s" % (NFS3_PROCEDURE_FSINFO, self.host))
			data, err = await self.nfs_request(NFS3_PROCEDURE_FSINFO, packer.get_buffer())
			if err is not None:
				raise err

			unpacker = nfs_pro_v3Unpacker(data)
			return unpacker.unpack_fsinfo3res(), None
		except Exception as e:
			return False, e

	async def pathconf(self, file_handle, auth=None):
		try:
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_fhandle3(file_handle)

			logger.debug("NFSv3 procedure %d: PATHCONF on %s" % (NFS3_PROCEDURE_PATHCONF, self.host))
			data, err = await self.nfs_request(NFS3_PROCEDURE_PATHCONF, packer.get_buffer())
			if err is not None:
				raise err

			unpacker = nfs_pro_v3Unpacker(data)
			return unpacker.unpack_pathconf3res(), None
		except Exception as e:
			return False, e


	async def commit(self, file_handle, count=0, offset=0, auth=None):
		try:
			file_handle = self.__handles[file_handle]
			packer = nfs_pro_v3Packer()
			packer.pack_commit3args(commit3args(file=nfs_fh3(file_handle), offset=offset, count=count))

			logger.debug("NFSv3 procedure %d: COMMIT on %s" % (NFS3_PROCEDURE_COMMIT, self.host))
			res = self.nfs_request(NFS3_PROCEDURE_COMMIT, packer.get_buffer())

			unpacker = nfs_pro_v3Unpacker(res)
			return unpacker.unpack_commit3res(), None
		except Exception as e:
			return False, e

async def amain():
	from anfs.protocol.rpc.common.target import RPCTarget
	from anfs.protocol.rpc.messages import AUTH_SYS
	from anfs.protocol.mount import Mount
	import pprint

	ip = '192.168.30.10'
	target = RPCTarget(ip)
	cred = AUTH_SYS(machinename='192.168.30.195') #192.168.30.195
	mount = Mount(target, cred)
	_, err = await mount.connect()
	if err is not None:
		raise err
	
	fhandle, err = await mount.mount('/volume1/NFSTEST')
	if err is not None:
		raise err
	print(fhandle)
	nfs = NFSv3(fhandle, target, cred)
	_, err = await nfs.connect()
	if err is not None:
		raise err
	res, err = await nfs.fsinfo(0)
	if err is not None:
		raise err
	print(res)

	res, err = await nfs.pathconf(0)
	if err is not None:
		raise err
	print(res)

	#_, err = await nfs.null()
	#if err is not None:
	#	raise err
	res, err = await nfs.access(0, 0x1f)
	async for entry, err in nfs.readdirplus(0):
		if err is not None:
			raise err
		print(str(entry))
	

	res, err = await nfs.getattr(0)
	print(res)

	res, err = await nfs.lookup(0, 'test')
	print(res)

	th = None
	async for entry, err in nfs.readdirplus(res.handle):
		if err is not None:
			raise err
		print(str(entry))
		if entry.name.startswith(b'Terasz-20180919-212510-1537385110.mp4'):
			th = entry.handle
			break
	
	print(th)
	res, err = await nfs.read(th, 0, 10)
	print(res)

def main():
	import asyncio
	asyncio.run(amain())
if __name__ == '__main__':
	main()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import asyncio
import traceback
import logging
import shlex
from tqdm import tqdm

from anfs import logger
from asysocks import logger as sockslogger
from asyauth import logger as authlogger
from anfs.examples.utils import PathCompleter
from anfs.external.aiocmd.aiocmd import aiocmd
from anfs.protocol.nfs3.common.factory import NFS3ConnectionFactory


class NFSClientConsole(aiocmd.PromptToolkitCmd):
	def __init__(self, url = None):
		aiocmd.PromptToolkitCmd.__init__(self, ignore_sigint=False)
		self.conn_url = url
		if url is not None and isinstance(url, NFS3ConnectionFactory) is False:
			self.conn_url = NFS3ConnectionFactory.from_url(url)
		self.connection = None
		self.mount = None
		self.__current_dir_handle = None
		self.__current_dirs = {}
		self.__current_files = {}
		self.__mounts = {}
		self.__current_mountpoint = None
		self.aliases['use'] = 'mount'
		

	async def do_connect(self, url = None):
		"""Performs connection and login"""
		try:			
			if self.conn_url is None and url is None:
				print('No URL was set, cant do logon')
			if url is not None and isinstance(url, NFS3ConnectionFactory) is False:
				self.conn_url = NFS3ConnectionFactory.from_url(url)

			logger.debug(self.conn_url.get_credential())
			logger.debug(self.conn_url.get_target())
			
			self.mount = self.conn_url.get_mount()
			_, err = await self.mount.connect()
			if err is not None:
				raise err
			
			logger.debug('Mount connect OK!')
			print('Connected! -mount-')
			
			return True
		except:
			traceback.print_exc()
			return False
		
	async def do_mounts(self):
		"""Lists all mountpoints"""
		
		if len(self.__mounts) == 0:
			mountres, err = await self.mount.export()
			if err is not None:
				raise err
			for mount in mountres:
				self.__mounts[mount.filesys] = mount
		
		for mount in self.__mounts:
			print(mount)
		
		return True

	async def do_mountinfo(self):
		"""Returns info about all mountpoints"""
		
		if len(self.__mounts) == 0:
			print('No mountpoints found! Did you run mounts?')
		
		for mount in self.__mounts:
			print(str(self.__mounts[mount]))
		
		return True
	
	def _mount_completions(self):
		return PathCompleter(get_current_dirs = lambda: list(self.__mounts.keys()))
	
	async def do_mount(self, mountpoint:str):
		"""Changes the current mountpoint"""

		if mountpoint == self.__current_mountpoint:
			print('Already using this mountpoint!')
			return True
		
		mhandle, err = await self.mount.mount(mountpoint)
		if err is not None:
			raise err
		
		self.connection = self.conn_url.get_client(mhandle)
		_, err = await self.connection.connect()
		if err is not None:
			raise err
		logger.debug('Connect OK!')
		print('Connected! -NFS-')
		self.prompt = '[%s]> ' % mountpoint
		self.__current_mountpoint = mountpoint
		self.__current_dir_handle = 0 #root is always 0
		await self.do_refreshcurrentdir()
		return True
		
	async def do_refreshcurrentdir(self):
		"""Refreshes the current directory"""
		if self.__current_dir_handle is None:
			print('Please use a mountpoint first!')
			return False
		
		self.__current_dirs = {}
		self.__current_files = {}
		async for fe, err in self.connection.readdirplus(self.__current_dir_handle):
			if err is not None:
				raise err
			if fe.type == 2:
				if fe.name == '.':
					continue
				self.__current_dirs[fe.name] = fe
			elif fe.type == 5:
				# symlink
				symlinkdata, err = await self.connection.readlink(fe.handle)
				if err is not None:
					raise err
				
				fe.symtarget = symlinkdata
				self.__current_files[fe.name] = fe
			else:
				self.__current_files[fe.name] = fe
		return True
	
	async def do_ls(self):
		"""Lists the current directory"""
		if self.__current_dir_handle is None:
			print('Please use a mountpoint first!')
			return False
		
		for d in self.__current_dirs:
			print(self.__current_dirs[d].to_line())
		for f in self.__current_files:
			print(self.__current_files[f].to_line())
		return True
	
	def get_current_dirs(self):
		if self.__current_dirs is None:
			return []
		curdirs = []
		for dirname in self.__current_dirs:
			if dirname.find(' ') != -1:
				dirname = "'%s'" % dirname
			curdirs.append(dirname)
		return curdirs
	
	def get_current_files(self):
		if self.__current_files is None:
			return []
		return list(self.__current_files.keys())
	
	def _cd_completions(self):
		return PathCompleter(get_current_dirs = self.get_current_dirs)
	
	async def do_cd(self, dirname:str):
		"""Changes the current directory"""
		if dirname not in self.__current_dirs:
			print('Directory not found!')
			return False
		
		self.__current_dir_handle = self.__current_dirs[dirname].handle
		self.prompt = '[%s%s]> ' % (self.__current_mountpoint, self.connection.handle_to_path(self.__current_dir_handle))
		await self.do_refreshcurrentdir()
		return True
	
	def _get_completions(self):
		return PathCompleter(get_current_dirs = self.get_current_files)
	
	async def do_get(self, filename:str):
		"""Downloads a file"""
		if filename not in self.__current_files:
			print('File not found!')
			return False
		
		fh = self.__current_files[filename].handle
		size = 0
		pbar = tqdm(total=self.__current_files[filename].size, unit='B', unit_scale=True)
		with open(filename, 'wb') as f:
			while size < self.__current_files[filename].size:
				data, err = await self.connection.read(fh, size, 10*1024*1024)
				if err is not None:
					raise err
				f.write(data)
				size += len(data)
				pbar.update(len(data))
		return True
	
	async def do_mkdir(self, dirname:str):
		"""Creates a directory"""
		if dirname in self.__current_dirs:
			print('Directory already exists!')
			return False
		
		entry, err = await self.connection.mkdir(self.__current_dir_handle, dirname)
		if err is not None:
			raise err
		if entry is not None and entry is not False:
			print('Directory created!')
		await self.do_refreshcurrentdir()
		return True
	
	def _rmdir_completions(self):
		return PathCompleter(get_current_dirs = self.get_current_dirs)
	
	async def do_rmdir(self, dirname:str):
		"""Deletes a directory"""
		if dirname not in self.__current_dirs:
			print('Directory not found!')
			return False
		
		entry, err = await self.connection.rmdir(self.__current_dir_handle, dirname)
		if err is not None:
			raise err
		if entry is not None and entry is not False:
			print('Directory deleted!')
		await self.do_refreshcurrentdir()
		return True
	
	def _rm_completions(self):
		return PathCompleter(get_current_dirs = self.get_current_files)
	
	async def do_rm(self, filename:str):
		"""Deletes a file"""
		if filename not in self.__current_files:
			print('File not found!')
			return False
		
		entry, err = await self.connection.remove(self.__current_dir_handle, filename)
		if err is not None:
			raise err
		if entry is not None and entry is not False:
			print('File deleted!')
		await self.do_refreshcurrentdir()
		return True
	
	async def do_touch(self, filename:str):
		"""Creates a file"""
		if filename in self.__current_files:
			print('File already exists!')
			return False
		
		entry, err = await self.connection.create(self.__current_dir_handle, filename, 1)
		if err is not None:
			raise err
		if entry is not None and entry is not False:
			print('File created!')
		await self.do_refreshcurrentdir()
		return True
	
	async def do_symlink(self, filename:str, linkname:str):
		"""Creates a symlink"""		
		entry, err = await self.connection.symlink(self.__current_dir_handle, filename, linkname)
		if err is not None:
			raise err
		
		if entry is not None and entry is not False:
			print('Symlink created!')
		await self.do_refreshcurrentdir()
		return True
	
	async def __getfile_internal(self, fileentry, dst_path):
		size = 0
		pbar = tqdm(total=fileentry.size, unit='B', unit_scale=True)
		with open(dst_path, 'wb') as f:
			while size < fileentry.size:
				data, err = await self.connection.read(fileentry.handle, size, 10*1024*1024)
				if err is not None:
					raise err
				f.write(data)
				size += len(data)
				pbar.update(len(data))
		return True, None
	
	async def __getdir_internal(self, direntry, dst_path):
		if not os.path.exists(dst_path):
			os.makedirs(dst_path)

		async for entry, err in self.connection.readdirplus(direntry.handle):
			if err is not None:
				raise err
			if entry.name in ('.', '..'):
				continue
			if entry.type not in [1,2]:
				print('Skipping %s for unknown type!' % entry.name)
				continue

			target_path = os.path.normpath(os.path.join(dst_path, entry.name))
			if not target_path.startswith(dst_path):
				print(f"Skipping {entry.name} due to path traversal attempt!")
				continue

			if entry.type == 2:
				await self.__getdir_internal(entry, target_path)
			else:
				await self.__getfile_internal(entry, target_path)
		return True, None
	
	async def do_getdir(self, dirname):
		try:
			if dirname not in self.__current_dirs:
				print('Directory not found!')
				return False
			# current path
			dstpath = os.path.join(os.getcwd(), dirname)
			return await self.__getdir_internal(self.__current_dirs[dirname], dstpath)
		except Exception as e:
			traceback.print_exc()
			return False

	async def do_services(self):
		try:
			portmap = self.conn_url.get_portmap()
			_, err = await portmap.connect()
			if err is not None:
				raise err
			programs, err = await portmap.dump()
			if err is not None:
				raise err
			
			for service in programs:
				print(service)
				await asyncio.sleep(0)
			return True
		except Exception as e:
			traceback.print_exc()
			return False

async def amain(args):
	client = NFSClientConsole(args.url)
	
	if len(args.commands) == 0:
		if args.no_interactive is True:
			print('Not starting interactive!')
			return
		res = await client._run_single_command('connect', [])
		if res is False:
			return
		await client.run()
	else:
		for command in args.commands:
			if command == 'i':
				await client.run()
				return
			cmd = shlex.split(command)
			res = await client._run_single_command(cmd[0], cmd[1:])
			if res is False:
				return

def main():
	import argparse
	
	parser = argparse.ArgumentParser(description='NFSv3 console')
	parser.add_argument('-v', '--verbose', action='count', default=0, help='Verbosity, can be stacked')
	parser.add_argument('-n', '--no-interactive', action='store_true')
	parser.add_argument('url', help='Connection string in URL format.')
	parser.add_argument('commands', nargs='*', help="Takes a series of commands which will be executed until error encountered. If the command is 'i' is encountered during execution it drops back to interactive shell.")

	args = parser.parse_args()
	
	
	###### VERBOSITY
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	else:
		sockslogger.setLevel(logging.DEBUG)
		logger.setLevel(logging.DEBUG)
		authlogger.setLevel(logging.DEBUG)
		logging.basicConfig(level=logging.DEBUG)


	asyncio.run(amain(args))


if __name__ == '__main__':
	main()



import io
import enum
from anfs.protocol.portmap.wellknown import wellknown_programs

class ipproto(enum.Enum):
    IPPROTO_TCP = 6
    IPPROTO_UDP = 17

class mapping:
    def __init__(self):
        self.program = None
        self.version = None
        self.protocol = None
        self.port = None
        self.name = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return mapping.from_buffer(io.BytesIO(data))
    
    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = mapping()
        result.program = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.version = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.protocol = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.port = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        if result.program in wellknown_programs:
            result.name = wellknown_programs[result.program]
        return result
    
    def to_bytes(self):
        result = self.program.to_bytes(4, byteorder='big', signed=False)
        result += self.version.to_bytes(4, byteorder='big', signed=False)
        result += self.protocol.to_bytes(4, byteorder='big', signed=False)
        result += self.port.to_bytes(4, byteorder='big', signed=False)
        return result
    
    def __str__(self):
        return 'Program: %s, Version: %s, Protocol: %s, Port: %s, Name: %s' % (self.program, self.version, ipproto(self.protocol).name, self.port, self.name)

class call_args:
    def __init__(self):
        self.program = None
        self.version = None
        self.procedure = None
        self.data = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return call_args.from_buffer(io.BytesIO(data))
    
    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = call_args()
        result.program = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.version = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.procedure = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        
        result.data = buff.read()
        return result
    
    def to_bytes(self):
        result = self.program.to_bytes(4, byteorder='big', signed=False)
        result += self.version.to_bytes(4, byteorder='big', signed=False)
        result += self.procedure.to_bytes(4, byteorder='big', signed=False)
        result += self.data
        return result

class call_result:
    def __init__(self):
        self.port = None
        self.data = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return call_result.from_buffer(io.BytesIO(data))
    
    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = call_result()
        result.port = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.data = buff.read()
        return result
    
    def to_bytes(self):
        result = self.port.to_bytes(4, byteorder='big', signed=False)
        result += self.data
        return result


import struct

from asysocks.unicomm.common.packetizers import Packetizer

# https://datatracker.ietf.org/doc/html/rfc1831

class RPCPacketizer(Packetizer):
	def __init__(self):
		Packetizer.__init__(self, 65535)
		self.in_buffer = b''
		self.__in_fragment = False
		self.fragments = []
	
	def process_buffer(self):
		preread = 4
		remaining_length = -1
		while True:
			if len(self.in_buffer) < preread:
				break
			lb = self.in_buffer[:preread]
			remaining_length = struct.unpack('!L', lb)[0]
			self.__in_fragment = not bool(remaining_length & 0x80000000)
			remaining_length &= 0x7fffffff
			if len(self.in_buffer) >= remaining_length+preread:
				data = self.in_buffer[4:remaining_length+preread]
				self.fragments.append(data)
				self.in_buffer = self.in_buffer[remaining_length+preread:]
				if not self.__in_fragment:
					yield b''.join(self.fragments)
					self.fragments = []
				continue
			break
		

	async def data_out(self, data):
		yield data

	async def data_in(self, data):
		if data is None:
			yield data
		self.in_buffer += data
		for packet in self.process_buffer():
			yield packet


import io
import enum
import time

# https://datatracker.ietf.org/doc/html/rfc4506

class auth_flavor(enum.Enum):
    AUTH_NONE       = 0
    AUTH_SYS        = 1
    AUTH_SHORT      = 2
    AUTH_DH         = 3
    AUTH_KERB       = 4 #/* kerberos auth, see RFC 2695 */
    AUTH_RSA        = 5 #/* RSA authentication */
    RPCSEC_GSS      = 6 #/* GSS-based RPC security for auth, integrity and privacy, RPC 5403 */

    AUTH_NW         = 30001           #NETWARE
    AUTH_SEC        = 200000          #TSIG NFS subcommittee
    AUTH_ESV        = 200004          #SVr4 ES

    AUTH_NQNFS      = 300000          #Univ. of Guelph - Not Quite NFS
    AUTH_GSSAPI     = 300001          #OpenVision <john.linn@ov.com>
    AUTH_ILU_UGEN   = 300002          #Xerox <janssen@parc.xerox.com> ILU Unsecured Generic Identity
    AUTH_SPNEGO     = 390000
                   #390000 - 390255 NFS 'pseudo' flavors for RPCSEC_GSS
                   #390003 - kerberos_v5 authentication, RFC 2623
                   #390004 - kerberos_v5 with data integrity, RFC 2623
                   #390005 - kerberos_v5 with data privacy, RFC 2623

    Reserved        = 200000000
    NEXT_INC        = 200100000

class opaque_auth:
    def __init__(self):
        self.flavor = None
        self.body = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return opaque_auth.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = opaque_auth()
        result.flavor = auth_flavor(int.from_bytes(buff.read(4), byteorder='big', signed=False))
        bodylen = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.body = buff.read(bodylen)
        return result
    
    def to_bytes(self):
        result = self.flavor.value.to_bytes(4, byteorder='big', signed=False)
        result += len(self.body).to_bytes(4, byteorder='big', signed=False)
        result += self.body
        return result
    

class msg_type(enum.Enum):
    CALL  = 0
    REPLY = 1

class reply_stat(enum.Enum):
    MSG_ACCEPTED = 0
    MSG_DENIED   = 1

class accept_stat(enum.Enum):
    SUCCESS      = 0
    PROG_UNAVAIL = 1
    PROG_MISMATCH= 2
    PROC_UNAVAIL = 3
    GARBAGE_ARGS = 4
    SYSTEM_ERR   = 5

class reject_stat(enum.Enum):
    RPC_MISMATCH = 0
    AUTH_ERROR   = 1

class auth_stat(enum.Enum):
    AUTH_OK      = 0
    AUTH_BADCRED      = 1,  #/* bad credential (seal broken)   */
    AUTH_REJECTEDCRED = 2,  #/* client must begin new session  */
    AUTH_BADVERF      = 3,  #/* bad verifier (seal broken)     */
    AUTH_REJECTEDVERF = 4,  #/* verifier expired or replayed   */
    AUTH_TOOWEAK      = 5,  #/* rejected for security reasons  */
    #* failed locally*/
    AUTH_INVALIDRESP  = 6,  #/* bogus response verifier        */
    AUTH_FAILED       = 7,  #/* reason unknown                 */
    #/* AUTH_KERB errors; deprecated.  See [RFC2695] */
    AUTH_KERB_GENERIC = 8,  #/* kerberos generic error */
    AUTH_TIMEEXPIRE = 9,    #/* time of credential expired */
    AUTH_TKT_FILE = 10,     #/* problem with ticket file */
    AUTH_DECODE = 11,       #/* can't decode authenticator */
    AUTH_NET_ADDR = 12,     #/* wrong net address in ticket */
    #/* RPCSEC_GSS GSS related errors*/
    RPCSEC_GSS_CREDPROBLEM = 13, #/* no credentials for user */
    RPCSEC_GSS_CTXPROBLEM = 14   #/* problem with context */

class rpc_msg:
    def __init__(self):
        self.xid = None
        self.msg_type = None
        self.body = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return rpc_msg.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = rpc_msg()
        result.xid = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.msg_type = msg_type(int.from_bytes(buff.read(4), byteorder='big', signed=False))
        if result.msg_type == msg_type.CALL:
            result.body = call_body.from_buffer(buff)
        else:
            result.body = reply_body.from_buffer(buff)
        return result
    
    def to_bytes(self):
        result = self.xid.to_bytes(4, byteorder='big', signed=False)
        result += self.msg_type.value.to_bytes(4, byteorder='big', signed=False)
        result += self.body.to_bytes()
        return result

class call_body:
    def __init__(self):
        self.rpcvers = 2
        self.prog = None
        self.vers = None
        self.proc = None
        self.cred = None
        self.verf = None
        self.data = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return call_body.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = call_body()
        result.rpcvers = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.prog = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.vers = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.proc = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.cred = opaque_auth.from_buffer(buff)
        result.verf = opaque_auth.from_buffer(buff)
        result.data = buff.read()
        return result
    
    def to_bytes(self):
        result = self.rpcvers.to_bytes(4, byteorder='big', signed=False)
        result += self.prog.to_bytes(4, byteorder='big', signed=False)
        result += self.vers.to_bytes(4, byteorder='big', signed=False)
        result += self.proc.to_bytes(4, byteorder='big', signed=False)
        result += self.cred.to_bytes()
        result += self.verf.to_bytes()
        result += self.data
        return result

class reply_body:
    def __init__(self):
        self.reply_stat = None
        self.reply = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return reply_body.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = reply_body()
        result.reply_stat = reply_stat(int.from_bytes(buff.read(4), byteorder='big', signed=False))
        if result.reply_stat == reply_stat.MSG_ACCEPTED:
            result.reply = accepted_reply.from_buffer(buff)
        else:
            result.reply = rejected_reply.from_buffer(buff)
        return result
    
    def to_bytes(self):
        result = self.reply_stat.to_bytes(4, byteorder='big', signed=False)
        result += self.reply.to_bytes()
        return result

class accepted_reply:
    def __init__(self):
        self.verf = None
        self.stat = None
        self.mismatch_info = None
        self.results = None

    @staticmethod
    def from_bytes(data: bytes):
        return accepted_reply.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = accepted_reply()
        result.verf = opaque_auth.from_buffer(buff)
        result.stat = accept_stat(int.from_bytes(buff.read(4), byteorder='big', signed=False))
        if result.stat == accept_stat.SUCCESS:
            result.results = buff.read()
        elif result.stat == accept_stat.PROG_MISMATCH:
            result.mismatch_info = mismatch_info.from_buffer(buff)
        return result
    
    def to_bytes(self):
        result = self.verf.to_bytes()
        result += self.stat.to_bytes(4, byteorder='big', signed=False)
        if self.stat == accept_stat.SUCCESS:
            result += self.results
        elif self.stat == accept_stat.PROG_MISMATCH:
            result += self.mismatch_info.to_bytes()
        return result

class rejected_reply:
    def __init__(self):
        self.stat = None
        self.mismatch_info = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return rejected_reply.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = rejected_reply()
        result.stat = reject_stat(int.from_bytes(buff.read(4), byteorder='big', signed=False))
        if result.stat == reject_stat.RPC_MISMATCH:
            result.mismatch_info = mismatch_info.from_buffer(buff)
        return result
    
    def to_bytes(self):
        result = self.stat.to_bytes(4, byteorder='big', signed=False)
        if self.stat == reject_stat.RPC_MISMATCH:
            result += self.mismatch_info.to_bytes()
        return result

class mismatch_info:
    def __init__(self):
        self.low = None
        self.high = None
    
    @staticmethod
    def from_bytes(data: bytes):
        return mismatch_info.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = mismatch_info()
        result.low = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.high = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        return result
    
    def to_bytes(self):
        result = self.low.to_bytes(4, byteorder='big', signed=False)
        result += self.high.to_bytes(4, byteorder='big', signed=False)
        return result

class authsys_params:
    def __init__(self, stamp = None, machinename = None, uid = None, gid = None, gids = None):
        self.stamp = stamp
        self.machinename = machinename
        self.uid = uid
        self.gid = gid
        self.gids = gids

        if self.stamp is None:
            self.stamp = int(time.time()) & 0xffff
        if self.machinename is None:
            self.machinename = 'anfs'
        if self.uid is None:
            self.uid = 0
        if self.gid is None:
            self.gid = 0
        if self.gids is None:
            self.gids = [0]
    
    @staticmethod
    def from_bytes(data: bytes):
        return authsys_params.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        result = authsys_params()
        result.stamp = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.machinename = buff.read(255)
        result.uid = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.gid = int.from_bytes(buff.read(4), byteorder='big', signed=False)
        result.gids = buff.read(16*4)
        return result
    
    def to_bytes(self):
        result = self.stamp.to_bytes(4, byteorder='big', signed=False)
        mn = self.machinename.encode('utf-8')
        result += len(mn).to_bytes(4, byteorder='big', signed=False)
        result += mn
        result += b'\x00'*((4-len(mn) % 4) % 4)
        result += self.uid.to_bytes(4, byteorder='big', signed=False)
        result += self.gid.to_bytes(4, byteorder='big', signed=False)
        if len(self.gids) == 1 and self.gids[0] == 0:
            result += b'\x00\x00\x00\x00'
        else:
            result += len(self.gids).to_bytes(4, byteorder='big', signed=False)
            for gid in self.gids:
                result += gid.to_bytes(4, byteorder='big', signed=False)
        return result
    
class AUTH_NULL(opaque_auth):
    def __init__(self):
        super().__init__()
        self.flavor = auth_flavor.AUTH_NONE
        self.body = b''

class AUTH_SYS(opaque_auth):
    def __init__(self, stamp = None, machinename = None, uid = None, gid = None, gids = None):
        super().__init__()
        params = authsys_params(stamp = stamp, machinename = machinename, uid = uid, gid = gid, gids = gids)
        self.flavor = auth_flavor.AUTH_SYS
        self.body = params.to_bytes()




class RPCException(Exception):
    pass

class RPCReplyException(RPCException):
    pass

class RPCReplyDeniedException(RPCReplyException):
    def __init__(self):
        super().__init__('RPC reply denied without reason provided.')

class RPCCallException(RPCException):
    pass


class RPCCallRejectedException(RPCCallException):
    def __init__(self, reject_code):
        self.code = reject_code
        self.msg = ''
        if(self.code == 1):
            self.msg = 'Remote procedure not exported by server'
        elif(self.code == 2):
            self.msg = 'Remote procedure version mismatch'
        elif(self.code == 3):
            self.msg = 'Remote procedure is not available'
        elif(self.code == 4):
            self.msg = 'Incorrect parameters'
        elif(self.code == 5):
            self.msg = 'System error. (remote)'
        super().__init__('RPC Call rejected. Code: %s, Message: %s' % (reject_code, self.msg))




import asyncio
import io
from typing import cast
from asysocks.unicomm.client import UniClient
from anfs.protocol.rpc.packetizer import RPCPacketizer
from anfs.protocol.rpc.common.exceptions import *
from anfs.protocol.rpc.messages import rpc_msg, msg_type, call_body, reply_body,\
	AUTH_NULL, reply_stat, accepted_reply, accept_stat
import traceback

# https://datatracker.ietf.org/doc/html/rfc5531
# https://github.com/CharmingYang0/NfsClient/

class RPC:
	def __init__(self, target, credential = AUTH_NULL()):
		self.target = target
		self.credential = credential
		self.connection = None
		self.__portmap = None
		self.__read_task = None
		self.__current_xid = 10
		self.__xid_table = {}
		if self.credential is None:
			self.credential = AUTH_NULL()

	async def __handle_incoming(self):
		try:
			async for message_data in self.connection.read():		
				message = rpc_msg.from_bytes(message_data)
				if message.msg_type == msg_type.REPLY:
					if message.xid in self.__xid_table:
						x = message.body
						await self.__xid_table[message.xid].put(message.body)
					else:
						print('Received a REPLY message, but no matching xid found. XID: %s' % message.xid)
				else:
					raise Exception('Received a CALL message, expected a REPLY message')
		except Exception as e:
			traceback.print_exc()
			await self.disconnect()

	def next_xid(self):
		t = self.__current_xid
		self.__current_xid += 1
		if self.__current_xid >= 0xffffffff:
			self.__current_xid = 10
		return t
	
	async def register_xid(self, xid):
		self.__xid_table[xid] = asyncio.Queue()
		return xid, self.__xid_table[xid]
	
	
	async def connect(self, program, version):
		try:
			if self.target.port == 0:
				protocol = self.target.get_rpcprotocol()
				# at this point we don't know the port, we need to query the portmapper
				if program is None or version is None or protocol is None:
					raise Exception('Program or port must be set!')
				from anfs.protocol.portmap import Portmap
				nt = self.target.get_newtarget(self.target.ip, 111)
				self.__portmap = Portmap(nt, self.credential)
				_, err = await self.__portmap.connect()
				if err is not None:
					return False, err
				port, err = await self.__portmap.getport(program, version, protocol)
				if err is not None:
					return False, err
				self.target = self.target.get_newtarget(self.target.ip, port)


			packetizer = RPCPacketizer()
			client = UniClient(self.target, packetizer)
			self.connection = await asyncio.wait_for(client.connect(), timeout=self.target.timeout)
			self.__read_task = asyncio.create_task(self.__handle_incoming())
			return True, None
		except Exception as e:
			return False, e

	async def disconnect(self):
		if self.connection is not None:
			await self.connection.close()
	
	async def __send(self, data):
		dsize = len(data)
		total_sent = 0
		data = io.BytesIO(data)
		while total_sent < dsize:
			fragment = data.read(self.target.fragsize)
			rpc_fragment_header = len(fragment)
			if rpc_fragment_header == 0:
				break
			if (rpc_fragment_header + total_sent) == dsize:
				rpc_fragment_header += 0x80000000
			total_sent += rpc_fragment_header
			
			fragment = rpc_fragment_header.to_bytes(4, byteorder='big', signed=False) + fragment
			await self.connection.write(fragment)
	
	async def call(self, program, version, procedure, data):
		xid, xid_queue = await self.register_xid(self.next_xid())
		try:
			call = rpc_msg()
			call.xid = xid
			call.msg_type = msg_type.CALL
			call.body = call_body()
			call.body.rpcvers = 2
			call.body.prog = program
			call.body.vers = version
			call.body.proc = procedure
			call.body.cred = self.credential
			call.body.verf = AUTH_NULL() # for both auth null and auth sys, the verifier is null
			call.body.data = data
			
			await self.__send(call.to_bytes())
			reply = await xid_queue.get()
			reply = cast(reply_body, reply)
			if reply.reply_stat != reply_stat.MSG_ACCEPTED:
				raise RPCReplyDeniedException()
			
			areply = cast(accepted_reply, reply.reply)

			if areply.stat != accept_stat.SUCCESS:
				raise RPCCallRejectedException(areply.stat.value)
			
			return areply.results, None
		except Exception as e:
			traceback.print_exc()
			return None, e
		finally:
			del self.__xid_table[xid]